import torch
from PIL import Image
from transformers import AutoModelForCausalLM, AutoProcessor, GenerationConfig
from typing import List, Union

MOLMO_MODELS = {
    'molmo-72b-0924': {
        'model': {'path': 'allenai/Molmo-72B-0924'},
        'processor': {'path': 'allenai/Molmo-72B-0924'},
    },
    'molmo-7b-d-0924': {
        'model': {'path': 'allenai/Molmo-7B-D-0924'},
        'processor': {'path': 'allenai/Molmo-7B-D-0924'},
    },
    'molmo-7b-o-0924': {
        'model': {'path': 'allenai/Molmo-7B-O-0924'},
        'processor': {'path': 'allenai/Molmo-7B-O-0924'},
    },
    'molmoe-1b-0924': {
        'model': {'path': 'allenai/MolmoE-1B-0924'},
        'processor': {'path': 'allenai/MolmoE-1B-0924'},
    },
}

class MOLMOVisionModel:
    video_mode = "concat"
    allows_image = True
    def __init__(self,
                 model_name='molmo-7b-d-0924',
                 device='cuda',
                 cache_dir=None):
        assert model_name in MOLMO_MODELS, f"Model {model_name} not found in MOLMO_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = MOLMO_MODELS[model_name]
        self.load_model()

    def load_model(self):
        model_config = self.model_info['model']
        self.model = AutoModelForCausalLM.from_pretrained(
            model_config['path'],
            trust_remote_code=True,
            torch_dtype='auto',
            device_map='auto',
            cache_dir=self.cache_dir
        )
        self.processor = AutoProcessor.from_pretrained(
            self.model_info['processor']['path'],
            trust_remote_code=True,
            torch_dtype='auto',
            device_map='auto',
            cache_dir=self.cache_dir
        )

    def load_images(self, paths: List[str]) -> List[Image.Image]:
        processed_data = []
        for path in paths:
            if path.startswith(("http://", "https://")):
                raise NotImplementedError("Web link image/video inputs are not yet supported for this model. Please use a local path, or otherwise, make a Github issue request if this feature is necessary.")
            if path.lower().endswith(('.mp4', '.avi', '.mov', '.mkv')):
                raise NotImplementedError("Video processing is not supported for MOLMO Vision model.")
            else:  # Regular image file
                image = Image.open(path).convert('RGB')
                processed_data.append(image)
        return processed_data

    def forward(self,
                images: List[str],
                texts: List[str],
                question_template: str = "Does this image show \"{}\"? Answer the question with Yes or No",
                answer_template: str = "Yes") -> torch.Tensor:
        assert len(paths) == len(texts), "Number of paths and texts must match"
        
        images = self.load_images(images)
        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]

        lm_probs = []
        for image, question, answer in zip(images, questions, answers):
            inputs = self.processor.process(images=[image], text=question)
            inputs = {k: v.to(self.model.device).unsqueeze(0) for k, v in inputs.items()}
            
            with torch.no_grad():
                outputs = self.model.generate_from_batch(
                    inputs,
                    GenerationConfig(
                        max_new_tokens=1,
                        return_dict_in_generate=True,
                        output_scores=True,
                        stop_strings="<|endoftext|>"
                    ),
                    tokenizer=self.processor.tokenizer
                )
            
            # Get the logits for the first (and only) generated token
            logits = outputs.scores[0]
            
            # Get the index of the "Yes" token
            yes_token_id = self.processor.tokenizer.encode(" " + answer)[0]

            # Apply softmax to get probabilities
            probs = torch.nn.functional.softmax(logits, dim=-1)
            
            # Get the probability for the "Yes" token
            yes_prob = probs[0, yes_token_id].item()
            lm_probs.append(yes_prob)
        
        return torch.tensor(lm_probs)
    def generate(self,
                paths: List[str],
                texts: List[str],
                max_new_tokens: int = 256) -> List[str]:
        """
        Generate text responses for the given images and prompts.
        
        Args:
            paths: List of paths to image files
            texts: List of prompt texts
            max_new_tokens: Maximum number of tokens to generate
            
        Returns:
            List of generated text responses
        """
        assert len(paths) == len(texts), "Number of paths and texts must match"
        
        images = self.load_images(paths)
        generated_texts = []
        

        

        for image, prompt in zip(images, texts):
            inputs = self.processor.process(images=[image], text=prompt)
            inputs = {k: v.to(self.model.device).unsqueeze(0) for k, v in inputs.items()}
            
            input_length = inputs['input_ids'].size(1)
            with torch.no_grad():
                outputs = self.model.generate_from_batch(
                    inputs,
                    GenerationConfig(
                        max_new_tokens=max_new_tokens,
                        return_dict_in_generate=True,
                        stop_strings="<|endoftext|>"
                    ),
                    tokenizer=self.processor.tokenizer
                )
            new_tokens = outputs.sequences[:, input_length:]
            
            # Decode the generated tokens
            text = self.processor.tokenizer.batch_decode(new_tokens, skip_special_tokens=True)
            generated_texts.extend(text)
            # generated_texts.append(text.strip())

        return generated_texts

    # def load_video(self, video_path, max_frames_num):
    #     raise NotImplementedError("Direct video processing is not supported for MOLMO Vision model.")

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)


"""Human fidelity module — VBench-2.0 dimension.

Assesses body, hand, and face quality in generated humans via pose
detection and landmark analysis.

Backend tiers:
  1. **DWPose** — Full-body + hand + face landmarks (dwpose / mmpose)
  2. **MediaPipe** — 33 body landmarks
  3. **Heuristic** — Haar cascades + skin segmentation
"""

import logging
from typing import Optional

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class HumanFidelityModule(PipelineModule):
    name = "human_fidelity"
    description = "Human body/hand/face fidelity (DWPose / MediaPipe / heuristic)"
    default_config = {}

    # MediaPipe landmark indices for anatomy checks
    _LIMB_PAIRS = [
        (11, 13, "left_upper_arm"),
        (12, 14, "right_upper_arm"),
        (13, 15, "left_forearm"),
        (14, 16, "right_forearm"),
        (23, 25, "left_thigh"),
        (24, 26, "right_thigh"),
        (25, 27, "left_shin"),
        (26, 28, "right_shin"),
    ]

    _SYMMETRY_PAIRS = [(0, 1), (2, 3), (4, 5), (6, 7)]

    def __init__(self, config=None):
        super().__init__(config)
        self._backend = "heuristic"
        self._mp_available = False
        self._dwpose_available = False
        self.mp_pose = None

    def setup(self) -> None:
        # Tier 1: DWPose
        try:
            from dwpose import DWposeDetector  # noqa: F401
            self._dwpose_available = True
            self._backend = "dwpose"
            logger.info("HumanFidelity using DWPose backend")
            return
        except ImportError:
            pass

        # Tier 2: MediaPipe
        try:
            import mediapipe as mp
            self.mp_pose = mp.solutions.pose
            self._mp_available = True
            self._backend = "mediapipe"
            logger.info("HumanFidelity using MediaPipe backend")
            return
        except ImportError:
            pass

        # Tier 3: Heuristic
        logger.info("HumanFidelity using heuristic backend (Haar + skin)")

    def process(self, sample: Sample) -> Sample:
        image = self._load_image(sample)
        if image is None:
            return sample

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        try:
            if self._backend == "dwpose":
                score, issues = self._compute_dwpose(image)
            elif self._backend == "mediapipe":
                score, issues = self._compute_mediapipe(image)
            else:
                score, issues = self._compute_heuristic(image)

            if score is not None:
                sample.quality_metrics.human_fidelity_score = score

            for issue in issues:
                sample.validation_issues.append(issue)

        except Exception as e:
            logger.warning("Human fidelity check failed: %s", e)

        return sample

    # ------------------------------------------------------------------ #
    # Tier 1: DWPose                                                       #
    # ------------------------------------------------------------------ #

    def _compute_dwpose(self, image: np.ndarray) -> tuple:
        from dwpose import DWposeDetector

        detector = DWposeDetector()
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        result = detector(image_rgb)
        issues = []

        if result is None or len(result.get("bodies", {}).get("candidate", [])) == 0:
            return None, issues

        # Body score from landmark confidence
        bodies = result.get("bodies", {})
        body_candidates = bodies.get("candidate", [])
        body_scores = [c[2] for c in body_candidates if len(c) > 2]
        body_score = float(np.mean(body_scores)) if body_scores else 0.5

        # Hand score
        hands = result.get("hands", [])
        if hands and len(hands) > 0:
            hand_scores = []
            for hand in hands:
                if hand is not None and len(hand) > 0:
                    hand_confs = [p[2] for p in hand if len(p) > 2]
                    if hand_confs:
                        hand_scores.append(float(np.mean(hand_confs)))
            hand_score = float(np.mean(hand_scores)) if hand_scores else 0.3
        else:
            hand_score = 0.3

        # Face score
        faces = result.get("faces", [])
        if faces and len(faces) > 0:
            face_scores = []
            for face in faces:
                if face is not None and len(face) > 0:
                    face_confs = [p[2] for p in face if len(p) > 2]
                    if face_confs:
                        face_scores.append(float(np.mean(face_confs)))
            face_score = float(np.mean(face_scores)) if face_scores else 0.3
        else:
            face_score = 0.3

        score = 0.4 * body_score + 0.3 * hand_score + 0.3 * face_score
        return float(np.clip(score, 0.0, 1.0)), issues

    # ------------------------------------------------------------------ #
    # Tier 2: MediaPipe                                                    #
    # ------------------------------------------------------------------ #

    def _compute_mediapipe(self, image: np.ndarray) -> tuple:
        issues = []
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        with self.mp_pose.Pose(
            static_image_mode=True,
            model_complexity=1,
            enable_segmentation=False,
            min_detection_confidence=0.5,
        ) as pose:
            results = pose.process(image_rgb)

            if not results.pose_landmarks:
                return None, issues

            landmarks = results.pose_landmarks.landmark
            h, w = image.shape[:2]

            # Body score: visibility × proportion correctness × symmetry
            avg_visibility = sum(lm.visibility for lm in landmarks) / len(landmarks)

            if avg_visibility < 0.3:
                issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.INFO,
                        message=f"Low pose visibility: {avg_visibility:.2f} (possible anatomy distortion)",
                        details={"avg_visibility": avg_visibility},
                    )
                )

            limb_lengths = self._compute_limb_lengths(landmarks, w, h)
            proportion_issues = self._check_proportions(limb_lengths)
            proportion_score = 1.0 - min(len(proportion_issues) * 0.2, 1.0)

            if proportion_issues:
                issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Abnormal limb proportions detected ({len(proportion_issues)} issues)",
                        details={"proportion_issues": proportion_issues},
                        recommendation="Generated human may have distorted body proportions.",
                    )
                )

            symmetry_ratio = self._check_symmetry(limb_lengths)
            symmetry_score = symmetry_ratio if symmetry_ratio is not None else 0.7

            if symmetry_ratio is not None and symmetry_ratio < 0.5:
                issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Poor bilateral symmetry: {symmetry_ratio:.2f}",
                        details={"symmetry_ratio": symmetry_ratio},
                        recommendation="Left/right limb lengths are highly asymmetric.",
                    )
                )

            body_score = (avg_visibility + proportion_score + symmetry_score) / 3.0

            # Hand score: check wrist/finger landmark visibility (indices 15-22)
            hand_indices = [15, 16, 17, 18, 19, 20, 21, 22]
            hand_vis = [landmarks[i].visibility for i in hand_indices if i < len(landmarks)]
            hand_score = float(np.mean(hand_vis)) if hand_vis else 0.3

            # Face score: check face landmark visibility (indices 0-10)
            face_indices = list(range(11))
            face_vis = [landmarks[i].visibility for i in face_indices if i < len(landmarks)]
            face_score = float(np.mean(face_vis)) if face_vis else 0.3

            score = 0.4 * body_score + 0.3 * hand_score + 0.3 * face_score
            return float(np.clip(score, 0.0, 1.0)), issues

    # ------------------------------------------------------------------ #
    # Tier 3: Heuristic                                                    #
    # ------------------------------------------------------------------ #

    def _compute_heuristic(self, image: np.ndarray) -> tuple:
        issues = []
        h, w = image.shape[:2]
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        # Face detection via Haar cascade
        face_cascade = cv2.CascadeClassifier(
            cv2.data.haarcascades + "haarcascade_frontalface_default.xml"
        )
        faces = face_cascade.detectMultiScale(gray, 1.1, 4)
        face_score = min(len(faces) * 0.5, 1.0) if len(faces) > 0 else 0.0

        # Skin detection via HSV range
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        lower_skin = np.array([0, 20, 70], dtype=np.uint8)
        upper_skin = np.array([20, 255, 255], dtype=np.uint8)
        skin_mask = cv2.inRange(hsv, lower_skin, upper_skin)
        skin_ratio = float(np.sum(skin_mask > 0)) / (h * w)

        # Body score based on skin presence
        body_score = min(skin_ratio * 5.0, 1.0)  # ~20% skin = 1.0

        # Hand region approximation (skin regions in lower-middle area)
        hand_region = skin_mask[h // 2:, w // 4: 3 * w // 4]
        hand_ratio = float(np.sum(hand_region > 0)) / (hand_region.size + 1)
        hand_score = min(hand_ratio * 8.0, 1.0)

        if len(faces) == 0 and skin_ratio < 0.05:
            # No human detected — return None
            return None, issues

        score = 0.4 * body_score + 0.3 * hand_score + 0.3 * face_score
        return float(np.clip(score, 0.0, 1.0)), issues

    # ------------------------------------------------------------------ #
    # Helpers (preserved from original)                                    #
    # ------------------------------------------------------------------ #

    def _compute_limb_lengths(self, landmarks, img_w: int, img_h: int) -> dict:
        lengths = {}
        for start_idx, end_idx, name in self._LIMB_PAIRS:
            lm_start = landmarks[start_idx]
            lm_end = landmarks[end_idx]
            if lm_start.visibility > 0.3 and lm_end.visibility > 0.3:
                dx = (lm_start.x - lm_end.x) * img_w
                dy = (lm_start.y - lm_end.y) * img_h
                lengths[name] = float(np.sqrt(dx * dx + dy * dy))
        return lengths

    def _check_proportions(self, limb_lengths: dict) -> list:
        issues = []
        ratio_checks = [
            ("left_upper_arm", "left_forearm", 0.6, 2.0, "left arm ratio"),
            ("right_upper_arm", "right_forearm", 0.6, 2.0, "right arm ratio"),
            ("left_thigh", "left_shin", 0.6, 1.8, "left leg ratio"),
            ("right_thigh", "right_shin", 0.6, 1.8, "right leg ratio"),
        ]
        for limb_a, limb_b, min_ratio, max_ratio, label in ratio_checks:
            if limb_a in limb_lengths and limb_b in limb_lengths:
                len_b = limb_lengths[limb_b]
                if len_b > 1e-6:
                    ratio = limb_lengths[limb_a] / len_b
                    if ratio < min_ratio or ratio > max_ratio:
                        issues.append(f"{label}: {ratio:.2f} (expected {min_ratio}-{max_ratio})")
        return issues

    def _check_symmetry(self, limb_lengths: dict) -> Optional[float]:
        ratios = []
        for left_idx, right_idx in self._SYMMETRY_PAIRS:
            left_name = self._LIMB_PAIRS[left_idx][2]
            right_name = self._LIMB_PAIRS[right_idx][2]
            if left_name in limb_lengths and right_name in limb_lengths:
                a, b = limb_lengths[left_name], limb_lengths[right_name]
                if max(a, b) > 1e-6:
                    ratios.append(min(a, b) / max(a, b))
        return float(np.mean(ratios)) if ratios else None

    def _load_image(self, sample: Sample) -> Optional[np.ndarray]:
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_count // 2)
                ret, frame = cap.read()
                cap.release()
                return frame if ret else None
            else:
                return cv2.imread(str(sample.path))
        except Exception:
            return None



from typing import List
import torch
import copy

from .vqa_model import VQAScoreModel
from .mm_utils import expand2square, load_pretrained_model, tokenizer_image_token
from ...constants import HF_CACHE_DIR, CONTEXT_LEN, SYSTEM_MSG, DEFAULT_IMAGE_TOKEN, IGNORE_INDEX
from .llava.model import LlavaLlamaForCausalLM, ModelArguments

default_question_template = 'Does this figure show "{}"? Please answer yes or no.'
default_answer_template = "Yes"

def format_question(question, conversation_style='chat'):
    if conversation_style == 'plain': # for 1st stage model
        question = DEFAULT_IMAGE_TOKEN + question
    elif conversation_style == 'chat': # for 2nd stage model
        question = SYSTEM_MSG + " USER: " + DEFAULT_IMAGE_TOKEN + "\n" + question + " ASSISTANT: "
    # elif conversation_style == 'chat_no_system': # for 2nd stage model
    #     question = "USER: " + DEFAULT_IMAGE_TOKEN + "\n" + question + " ASSISTANT: "
    # elif conversation_style == 'chat_ood_system': # for 2nd stage model
    #     question = SYSTEM_MSG + " HUMAN: " + DEFAULT_IMAGE_TOKEN + "\n" + question + " GPT: "
    # elif conversation_style == 'chat_swap_image_question': # for 2nd stage model
    #     question = SYSTEM_MSG + " HUMAN: " + question + "\n" + DEFAULT_IMAGE_TOKEN + " GPT: "
    else:
        raise NotImplementedError()
    return question

def format_answer(answer, conversation_style='chat'):
    if conversation_style == 'plain': # for 1st stage model
        answer = answer + "\n"
    elif conversation_style == 'chat': # for 2nd stage model
        answer = answer + "</s>"
    # elif conversation_style in ['chat_no_system', 'chat_ood_system', 'chat_swap_image_question']: # for 2nd stage model
    #     answer = answer + "</s>"
    else:
        raise NotImplementedError()
    return answer

LLAVA_MODELS = {
    # llava-v1.6
    # 'llava-v1.6-13b': {
    #     'tokenizer' : {
    #         'path': 'liuhaotian/llava-v1.6-vicuna-13b',
    #     },
    #     'model': {
    #         'path': 'liuhaotian/llava-v1.6-vicuna-13b',
    #         'conversation': 'chat',
    #         'image_aspect_ratio': 'pad',
    #     },
    # },
    # We recommend using 'llava-v1.5-13b' for maximal performance.
    # If you want to use a smaller model, we recommend using 'llava-v1.5-7b'.
    'llava-v1.5-13b': {
        'tokenizer' : {
            'path': 'liuhaotian/llava-v1.5-13b',
        },
        'model': {
            'path': 'liuhaotian/llava-v1.5-13b',
            'conversation': 'chat',
            'image_aspect_ratio': 'pad',
        },
    },
    'llava-v1.5-7b': {
        'tokenizer' : {
            'path': 'liuhaotian/llava-v1.5-7b',
        },
        'model': {
            'path': 'liuhaotian/llava-v1.5-7b',
            'conversation': 'chat',
            'image_aspect_ratio': 'pad',
        },
    },
    # The following models are suboptimal, but are included for completeness.
    # 'llava-v1.5-13b-no-system': {
    #     'tokenizer' : {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #     },
    #     'model': {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #         'conversation': 'chat_no_system',
    #         'image_aspect_ratio': 'pad',
    #     },
    # },
    # 'llava-v1.5-13b-ood-system': {
    #     'tokenizer' : {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #     },
    #     'model': {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #         'conversation': 'chat_ood_system',
    #         'image_aspect_ratio': 'pad',
    #     },
    # },
    # 'llava-v1.5-13b-swap': {
    #     'tokenizer' : {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #     },
    #     'model': {
    #         'path': 'liuhaotian/llava-v1.5-13b',
    #         'conversation': 'chat_swap_image_question',
    #         'image_aspect_ratio': 'pad',
    #     },
    # },
    # 'llava-v1.5-13b-stage-1': {
    #     'tokenizer' : {
    #         'path': 'lmsys/vicuna-13b-v1.5',
    #         'model_max_length': CONTEXT_LEN,
    #         'padding_side': 'right',
    #     },
    #     'model': {
    #         'path': 'lmsys/vicuna-13b-v1.5',
    #         'mmprojector_repo': 'liuhaotian/llava-v1.5-mlp2x-336px-pretrain-vicuna-13b-v1.5',
    #         'mmprojector_name': 'mm_projector.bin',
    #         'conversation': "plain",
    #         'image_aspect_ratio': 'square',
    #     },
    # },
    # 'llava-v1.5-7b-stage-1': {
    #     'tokenizer' : {
    #         'path': 'lmsys/vicuna-7b-v1.5',
    #         'model_max_length': CONTEXT_LEN,
    #         'padding_side': 'right',
    #     },
    #     'model': {
    #         'path': 'lmsys/vicuna-7b-v1.5',
    #         'mmprojector_repo': 'liuhaotian/llava-v1.5-mlp2x-336px-pretrain-vicuna-7b-v1.5',
    #         'mmprojector_name': 'mm_projector.bin',
    #         'conversation': "plain",
    #         'image_aspect_ratio': 'square',
    #     },
    # },
    # The following models are built on top of LLaVA-1.5 and integrate well with LLaVA-1.5 codebase
    'sharegpt4v-7b': {
        'tokenizer' : {
            'path': 'Lin-Chen/ShareGPT4V-7B',
        },
        'model': {
            'path': 'Lin-Chen/ShareGPT4V-7B',
            'conversation': 'chat',
            'image_aspect_ratio': 'pad',
        },
    },
    'sharegpt4v-13b': {
        'tokenizer' : {
            'path': 'Lin-Chen/ShareGPT4V-13B',
        },
        'model': {
            'path': 'Lin-Chen/ShareGPT4V-13B',
            'conversation': 'chat',
            'image_aspect_ratio': 'pad',
        },
    },
}


class LLaVAModel(VQAScoreModel):
    """A wrapper for the LLaVA-1.5 models"""
    video_mode = "concat"
    allows_image = True
    def __init__(self,
                 model_name='llava-v1.5-13b',
                 device='cuda',
                 cache_dir=HF_CACHE_DIR):
        assert model_name in LLAVA_MODELS
        
        super().__init__(model_name=model_name,
                         device=device,
                         cache_dir=cache_dir)

    def load_model(self):
        """Load the model, tokenizer, image transform
        """
        model_args = ModelArguments()
        model_max_length = LLAVA_MODELS[self.model_name]['tokenizer']['model_max_length'] \
            if 'model_max_length' in LLAVA_MODELS[self.model_name]['tokenizer'] else None
        padding_side = LLAVA_MODELS[self.model_name]['tokenizer']['padding_side'] \
            if 'padding_side' in LLAVA_MODELS[self.model_name]['tokenizer'] else None
        mmprojector_repo = LLAVA_MODELS[self.model_name]['model']['mmprojector_repo'] \
            if 'mmprojector_repo' in LLAVA_MODELS[self.model_name]['model'] else None
        mmprojector_name = LLAVA_MODELS[self.model_name]['model']['mmprojector_name'] \
            if 'mmprojector_name' in LLAVA_MODELS[self.model_name]['model'] else None
        
        # default is 'pad' (llava-1.5 says this reduces hallucination)
        # stage-1 models use 'square'
        self.image_aspect_ratio = LLAVA_MODELS[self.model_name]['model']['image_aspect_ratio'] \
            if 'image_aspect_ratio' in LLAVA_MODELS[self.model_name]['model'] else 'pad'
        
        self.conversational_style = LLAVA_MODELS[self.model_name]['model']['conversation']
        
        self.context_len = CONTEXT_LEN
        
        self.tokenizer, self.model, self.image_processor = load_pretrained_model(
            LlavaLlamaForCausalLM,
            model_args,
            model_path=LLAVA_MODELS[self.model_name]['model']['path'],
            tokenizer_path=LLAVA_MODELS[self.model_name]['tokenizer']['path'],
            model_max_length=model_max_length,
            padding_side=padding_side,
            image_aspect_ratio=self.image_aspect_ratio,
            mmprojector_repo=mmprojector_repo,
            mmprojector_name=mmprojector_name,
            device=self.device,
            cache_dir=self.cache_dir
        )
        self.tokenizer.pad_token = self.tokenizer.unk_token

    def load_images(self,
                    image: List[str]) -> torch.Tensor:
        """Load the image(s), and return a tensor (after preprocessing) put on self.device
        """
        if any(path.startswith(("http://", "https://")) for path in image):
            raise NotImplementedError("Web link image/video inputs are not yet supported for this model. Please use a local path, or otherwise, make a Github issue request if this feature is necessary.")
        image = [self.image_loader(x) for x in image]
        if self.image_aspect_ratio == 'pad':
            image = [expand2square(image, tuple(int(x*255) for x in self.image_processor.image_mean)) for image in image]
        image = [self.image_processor.preprocess(image, return_tensors='pt')['pixel_values'][0] for image in image]
        assert all(x.shape == image[0].shape for x in image)
        image = torch.stack(image, dim=0).to(self.device)
        return image

    @torch.no_grad()
    @torch.autocast(device_type='cuda', dtype=torch.bfloat16)
    def forward(self,
                images: List[str],
                texts: List[str],
                question_template: str=default_question_template,
                answer_template: str=default_answer_template) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        assert len(images) == len(texts), "Number of images and texts must match"
        # Turn "a photo of a dog" into
        # Q: "Does this figure show "a photo of a dog"? Please answer yes or no."
        # A: "Yes"
        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]
        
        # Formatting for LLaVA-1.5 desired input including system message and image tokens
        questions = [format_question(question, conversation_style=self.conversational_style) for question in questions]
        answers = [format_answer(answer, conversation_style=self.conversational_style) for answer in answers]
        
        images = self.load_images(images)
        
        prompts = [qs + ans for qs, ans in zip(questions, answers)]
        
        input_ids = [tokenizer_image_token(prompt, self.tokenizer, return_tensors='pt') for prompt in prompts]
        labels = copy.deepcopy(input_ids)
        for label, qs in zip(labels, questions):
            tokenized_len = len(tokenizer_image_token(qs, self.tokenizer))
            if qs[-1] == " ":
                tokenized_len -= 1 # because white space
            label[:tokenized_len] = IGNORE_INDEX
    
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids,
            batch_first=True,
            padding_value=self.tokenizer.pad_token_id)
        labels = torch.nn.utils.rnn.pad_sequence(labels,
                                                batch_first=True,
                                                padding_value=IGNORE_INDEX)
        input_ids = input_ids[:, :self.tokenizer.model_max_length]
        labels = labels[:, :self.tokenizer.model_max_length]
            
        attention_mask = input_ids.ne(self.tokenizer.pad_token_id)
        input_ids, attention_mask, labels = input_ids.to(self.device), attention_mask.to(self.device), labels.to(self.device)
        input_ids, attention_mask, past_key_values, inputs_embeds, labels = self.model.prepare_inputs_labels_for_multimodal(
            input_ids,
            attention_mask,
            None,
            labels,
            images
        )
        
        assert input_ids is None, "input_ids should be None for LLaVA-1.5"
        assert past_key_values is None, "past_key_values should be None for LLaVA-1.5"
        model_input_kwargs = {
            'input_ids': input_ids, # None for LLaVA-1.5
            'attention_mask': attention_mask,
            'past_key_values': past_key_values,
            'inputs_embeds': inputs_embeds,
            'use_cache': None,
            'output_attentions': None,
            'output_hidden_states': None,
            'return_dict': False,
        }
        
        outputs = self.model.model(
            **model_input_kwargs
        )

        hidden_states = outputs[0]
        logits = self.model.lm_head(hidden_states)

        # Shift so that tokens < n predict n
        shift_logits = logits[..., :-1, :].contiguous()
        shift_labels = labels[..., 1:].contiguous()
        # Flatten the tokens
        loss_fct = torch.nn.CrossEntropyLoss(reduction='mean')
        shift_labels = shift_labels.to(shift_logits.device)
        lm_prob = torch.zeros(shift_logits.shape[0])
        for k in range(lm_prob.shape[0]):
            lm_prob[k] = (-loss_fct(shift_logits[k], shift_labels[k])).exp()
        return lm_prob
    @torch.no_grad()
    @torch.autocast(device_type='cuda', dtype=torch.bfloat16)
    def generate(self,
                images: List[str],
                texts: List[str],
                max_new_tokens: int = 256) -> List[str]:
        assert len(images) == len(texts), "Number of images and texts must match"
        
        questions = texts
        questions = [format_question(question, conversation_style=self.conversational_style) for question in questions]
        
        images = self.load_images(images)
        
        input_ids = [tokenizer_image_token(prompt, self.tokenizer, return_tensors='pt') for prompt in questions]
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids,
            batch_first=True,
            padding_value=self.tokenizer.pad_token_id)
        input_ids = input_ids[:, :self.tokenizer.model_max_length]
            
        attention_mask = input_ids.ne(self.tokenizer.pad_token_id)
        input_ids, attention_mask = input_ids.to(self.device), attention_mask.to(self.device)
        input_ids, attention_mask, past_key_values, inputs_embeds, _ = self.model.prepare_inputs_labels_for_multimodal(
            input_ids,
            attention_mask,
            None,
            None,
            images
        )
        
        model_input_kwargs = {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'past_key_values': past_key_values,
            'inputs_embeds': inputs_embeds,
            'max_new_tokens': max_new_tokens
        }
        
        outputs = self.model.generate(**model_input_kwargs)
        generated_texts = [self.tokenizer.decode(output, skip_special_tokens=True).strip() for output in outputs]
        
        return generated_texts


# Copyright (c) Meta Platforms, Inc. and affiliates.

import math
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Tuple, Union

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.attention.flex_attention import (BlockMask, _mask_mod_signature,
                                               flex_attention)
from xformers.ops import AttentionBias, fmha

from ..core import probe


class InitStdFactor(Enum):
    DISABLED = "disabled"  # Init std is divided by 1.0
    GLOBAL_DEPTH = "global_depth"  # Init std is divided by sqrt(2*n_layers)
    CURRENT_DEPTH = "current_depth"  # Init std is divided by sqrt(2*depth)
    DIM_RATIO = "dim_ratio"  # Init std is divided by model_dim/4096


@dataclass
class BaseTransformerArgs:
    dim: int = 512
    n_layers: int = 8
    head_dim: Optional[int] = None
    n_heads: Optional[int] = None
    n_kv_heads: Optional[int] = None

    ffn_dim_multiplier: Optional[float] = None

    multiple_of: int = 256

    norm_eps: float = 1e-5

    rope_theta: float = 10000.0

    old_context_len: int = 8192
    rope_scale_factor: int = 1
    low_freq_factor: int = 1
    high_freq_factor: int = 32

    init_base_std: Optional[float] = None
    init_std_factor: str = "disabled"

    max_seqlen: int = 1024


def cross_entropy(pred, target, **kwargs):
    return F.nll_loss(
        F.log_softmax(pred.flatten(end_dim=-2).float(), -1),
        target.flatten(end_dim=-1),
        **kwargs,
    )


def repeat_kv(x: torch.Tensor, n_rep: int, dim: int) -> torch.Tensor:
    """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
    assert dim == 2, "Only dim=2 is supported. Check the implementation for other dims."
    bs, slen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        x[:, :, :, None, :]
        .expand(bs, slen, n_kv_heads, n_rep, head_dim)
        .reshape(bs, slen, n_kv_heads * n_rep, head_dim)
    )


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor, seq_dim: int):
    """
    Reshape frequency tensor for broadcasting it with another tensor.

    This function reshapes the frequency tensor to have the same shape as the target tensor 'x'
    for the purpose of broadcasting the frequency tensor during element-wise operations.

    Args:
        freqs_cis (torch.Tensor): Frequency tensor to be reshaped.
        x (torch.Tensor): Target tensor for broadcasting compatibility.
        seq_dim (int): Sequence dimension index.

    Returns:
        torch.Tensor: Reshaped frequency tensor.
    """
    ndim = x.ndim
    assert 0 <= seq_dim < ndim
    assert freqs_cis.shape == (
        x.shape[seq_dim],
        x.shape[-3],
        2,
        2,
    ), f"freqs_cis vs x: {(freqs_cis.shape, x.shape)}"
    shape = [
        d if i == seq_dim or i == ndim - 3 else 1 for i, d in enumerate(x.shape[:-2])
    ] + [2, 2]
    return freqs_cis.view(*shape)


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    seq_dim: int,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    xq_ = xq.reshape(*xq.shape[:-1], -1, 1, 2)  # B S H D -> B S H D/2 1 2
    xk_ = xk.reshape(*xk.shape[:-1], -1, 1, 2)  # B S H D -> B S H D/2 1 2
    freqs_cis = reshape_for_broadcast(
        freqs_cis, xq_, seq_dim
    ).float()  # S D/2 2 2 -> 1 S 1 D/2 2 2
    xq_out = (xq_ * freqs_cis).sum(5).flatten(3)
    xk_out = (xk_ * freqs_cis).sum(5).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)


def causal_mask(b, h, q_idx, kv_idx):
    return q_idx >= kv_idx


def lengths_to_start_ids(lengths):
    doc_start = lengths.cumsum(0)
    doc_start = doc_start.roll(1)
    doc_start[0] = 0
    return doc_start


def lengths_to_local_ids(lengths):
    assert lengths.ndim == 1
    nb_seqs = lengths.size(0)
    total_seqlen = lengths.sum()
    # This gives the document id of each token
    doc_id = torch.repeat_interleave(lengths)
    # Compute document start for each document
    doc_start = lengths_to_start_ids(lengths)
    # Compute document start for each token
    doc_start = doc_start[doc_id]
    # Compute the position of each token within each document
    tok_id = torch.arange(total_seqlen, device=lengths.device) - doc_start

    return doc_id, tok_id


def generate_doc_mask_mod(
    mask_mod: _mask_mod_signature,
    lengths: torch.Tensor,
    kv_lengths: Optional[torch.Tensor] = None,
) -> _mask_mod_signature:
    """Generates mask mods that apply to inputs to flex attention in the sequence stacked
    format.

    Args:
        mask_mod: The mask mod to apply to the documents
        lengths: Lengths of each document

    Note:
        What is the sequence stacked format? When assembling batches of inputs, we
        take multiple sequences and stack them together to form 1 large sequence. We then
        use masking to ensure that the attention scores are only applied to tokens within
        the same document.

    Example:

    - Square mask
      doc_mask         lengths
      a a b b b c c    2 3 2
    a 1 0 0 0 0 0 0
    a 1 1 0 0 0 0 0
    b 0 0 1 0 0 0 0
    b 0 0 1 1 0 0 0
    b 0 0 1 1 1 0 0
    c 0 0 0 0 0 1 0
    c 0 0 0 0 0 1 1

    """
    kv_lengths = kv_lengths if kv_lengths is not None else lengths
    q_document_id, q_token_id = lengths_to_local_ids(lengths)
    kv_document_id, kv_token_id = lengths_to_local_ids(kv_lengths)
    q_max_idx = lengths.sum() - 1
    kv_max_idx = kv_lengths.sum() - 1

    def doc_mask_mod(b, h, q_idx, kv_idx):
        q_idx_cap = torch.minimum(q_max_idx, q_idx)
        kv_idx_cap = torch.minimum(kv_max_idx, kv_idx)
        valid_idx = (q_idx <= q_max_idx) & (kv_idx <= kv_max_idx)
        same_doc = q_document_id[q_idx_cap] == kv_document_id[kv_idx_cap]
        q_logical = q_token_id[q_idx_cap]
        kv_logical = kv_token_id[kv_idx_cap]
        inner_mask = mask_mod(b, h, q_logical, kv_logical)
        return same_doc & inner_mask & valid_idx

    return doc_mask_mod


# Rotary embedding as in xformer, see if torchtrain implementation is not better. Also might be usefull to make it work with batch*seqlen collapsed.
class RotaryEmbedding(torch.nn.Module):
    """
    RotaryEmbedding Module
    """

    def __init__(
        self,
        theta: float,
        head_dim: int,
        max_seqlen: int = 1024,
        scale_factor: int = 1,
        low_freq_factor: int = 1,
        high_freq_factor: int = 32,
        old_context_len: int = 8192,
    ):
        super().__init__()

        self.theta = theta
        self.head_dim = head_dim
        self.max_seqlen = max_seqlen
        self.scale_factor = scale_factor
        self.low_freq_factor = low_freq_factor
        self.high_freq_factor = high_freq_factor
        self.old_context_len = old_context_len
        if scale_factor != 1:
            self.low_freq_wavelen = old_context_len / low_freq_factor
            self.high_freq_wavelen = old_context_len / high_freq_factor
            assert self.low_freq_wavelen >= self.high_freq_wavelen

    def reset_parameters(self):
        self.register_buffer(
            "freqs_cis",
            self.precompute_freqs_cis(
                dim=self.head_dim, end=self.max_seqlen, theta=self.theta
            ),
            persistent=False,
        )

    def apply_scaling(self, freqs):
        if self.scale_factor == 1:
            return freqs
        new_freqs = []
        for freq in freqs:
            wavelen = 2 * math.pi / freq
            if wavelen < self.high_freq_wavelen:
                new_freqs.append(freq)
            elif wavelen > self.low_freq_wavelen:
                new_freqs.append(freq / self.scale_factor)
            else:
                assert self.low_freq_wavelen != self.high_freq_wavelen
                smooth = (self.old_context_len / wavelen - self.low_freq_factor) / (
                    self.high_freq_factor - self.low_freq_factor
                )
                new_freqs.append(
                    (1 - smooth) * freq / self.scale_factor + smooth * freq
                )
        return torch.tensor(new_freqs, dtype=freqs.dtype, device=freqs.device)

    def precompute_freqs_cis(
        self,
        dim: int,
        end: int,
        theta: float = 10000.0,
    ):
        """
        Precompute the frequency tensor for complex exponentials (cis) with given dimensions.

        This function calculates a frequency tensor with complex exponentials using the given dimension 'dim'
        and the end index 'end'. The 'theta' parameter scales the frequencies.
        The returned tensor contains complex values in complex64 data type.

        Args:
            dim (int): Dimension of the frequency tensor.
            end (int): End index for precomputing frequencies.
            theta (float, optional): Scaling factor for frequency computation. Defaults to 10000.0.

        Returns:
            torch.Tensor: Precomputed frequency tensor with complex exponentials.
        """
        freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
        freqs = self.apply_scaling(freqs)

        t = torch.arange(end, device=freqs.device)
        freqs = torch.outer(t, freqs).float()

        cos, sin = freqs.cos(), freqs.sin()

        return torch.stack((cos, -sin, sin, cos), dim=-1).view(*freqs.size(), 2, 2)

    def forward(
        self, seqlen: Optional[int] = None, tok_idx: Optional[torch.Tensor] = None
    ):
        """
        Return freqs_cis corresponding to consecutive seqlen positions or the corresponding tok_idx positions
        Args:
            seqlen (int): Contiguous sequence length
            tok_idx (torch.Tensor[int]): Position indices of each token this overrides seqlen

        Returns:
            Tuple(torch.Tensor, torch.Tensor): Embedded input tensor and freqs_cis
        """
        test = (seqlen is not None) or (tok_idx is not None)
        assert test, "Should provide atleast seqlen or tok_idx"
        if tok_idx is not None:
            return self.freqs_cis[tok_idx]
        elif seqlen is not None:
            return self.freqs_cis[0:seqlen]


class RMSNorm(nn.Module):
    """
    Initialize the RMSNorm normalization layer.

    Args:
        dim (int): The dimension of the input tensor.
        eps (float, optional): A small value added to the denominator for numerical stability. Default is 1e-6.

    Attributes:
        eps (float): A small value added to the denominator for numerical stability.
        weight (nn.Parameter): Learnable scaling parameter.

    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x: torch.Tensor):
        return x * torch.rsqrt((x * x).mean(-1, keepdim=True) + self.eps)

    def forward(self, x: torch.Tensor):
        x = probe.log_stats(x, "resid")
        output = self._norm(x.float())
        return (output * self.weight.float()).type_as(x)

    def reset_parameters(self):
        torch.nn.init.ones_(self.weight)  # type: ignore


class TiedLinear(nn.Module):
    def __init__(self, tied_module: nn.Module) -> None:
        super().__init__()
        self.tied_module = tied_module
        if not hasattr(tied_module, "weight"):
            raise AttributeError(
                "Provided module does not have attribute 'weight'. Please check your tied_module."
            )

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(x, self.tied_module.weight)


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        head_dim: int,
        n_heads: int,
        n_kv_heads: int,
        rope_theta: float,
    ):
        super().__init__()

        self.dim = dim
        self.head_dim = head_dim
        self.rope_theta = rope_theta

        self.n_heads = n_heads
        self.n_kv_heads = n_kv_heads
        self.heads_per_group = self.n_heads // self.n_kv_heads

        self.wq = nn.Linear(
            dim,
            n_heads * head_dim,
            bias=False,
        )
        self.wk = nn.Linear(
            dim,
            n_kv_heads * head_dim,
            bias=False,
        )
        self.wv = nn.Linear(
            dim,
            n_kv_heads * head_dim,
            bias=False,
        )

        self.wo = nn.Linear(
            n_heads * head_dim,
            dim,
            bias=False,
        )

    def forward(
        self,
        x: torch.Tensor,
        freq_cis: torch.Tensor,
        tok_idx: Optional[torch.Tensor] = None,
        mask: Optional[Union[BlockMask, AttentionBias, str]] = None,
        attn_impl: str = "sdpa",
    ) -> torch.Tensor:
        # B S D
        bsz, seq_len, dim = x.shape
        xq = self.wq(x.view_as(x))
        xk = self.wk(x.view_as(x))
        xv = self.wv(x.view_as(x))

        output_shape = xq.shape
        # B S D -> B S H D
        xq = xq.view(bsz, seq_len, self.n_heads, self.head_dim)
        xk = xk.view(bsz, seq_len, self.n_kv_heads, self.head_dim)
        xv = xv.view(bsz, seq_len, self.n_kv_heads, self.head_dim)

        xq, xk = apply_rotary_emb(xq, xk, 1, freq_cis[0:seq_len])

        # This condition helps us be easily compatible
        # with inference by adding a pluggable KVCache
        if hasattr(self, "kv_cache"):
            xk, xv = self.kv_cache.update(xk, xv, tok_idx)

        xk = repeat_kv(xk, self.heads_per_group, dim=2)
        xv = repeat_kv(xv, self.heads_per_group, dim=2)

        if attn_impl == "flex_attention":
            assert mask is None or isinstance(mask, BlockMask)
            xq, xk, xv = map(lambda e: e.transpose(1, 2), (xq, xk, xv))
            output = flex_attention(xq, xk, xv, block_mask=mask)
            output = output.transpose(1, 2).contiguous()  # B H S D -> B S H D

        elif attn_impl == "fmha":
            assert mask is None or isinstance(mask, AttentionBias)
            output = fmha.memory_efficient_attention(xq, xk, xv, attn_bias=mask)
            # This uses B S H D instead of B H S D of pytorch

        elif attn_impl == "sdpa":
            xq, xk, xv = map(lambda e: e.transpose(1, 2), (xq, xk, xv))
            assert mask is None or isinstance(mask, (str, torch.Tensor))
            is_causal = (mask == "causal") if isinstance(mask, str) else False
            mask = mask if isinstance(mask, torch.Tensor) else None
            output = F.scaled_dot_product_attention(
                xq,
                xk,
                xv,
                is_causal=is_causal,
                attn_mask=mask,
            )
            output = output.transpose(1, 2).contiguous()  # B H S D -> B S H D
        else:
            raise NotImplementedError(
                f"Attention implementation {attn_impl} not supported"
            )

        output = self.wo(output.reshape(output_shape))

        return output

    def reset_parameters(self, init_std=None, factor=1.0):
        init_std = init_std or (self.dim ** (-0.5))

        for w in [self.wq, self.wk, self.wv]:
            nn.init.trunc_normal_(
                w.weight,
                mean=0.0,
                std=init_std,
                a=-3 * init_std,
                b=3 * init_std,
            )

        nn.init.trunc_normal_(
            self.wo.weight,
            mean=0.0,
            std=init_std / factor,
            a=-3 * init_std,
            b=3 * init_std,
        )


class FeedForward(nn.Module):
    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        multiple_of: int,
        ffn_dim_multiplier: Optional[float],
        mp_size: int = 1,
    ):
        super().__init__()

        hidden_dim = int(2 * hidden_dim / 3)
        if ffn_dim_multiplier is not None:
            hidden_dim = int(ffn_dim_multiplier * hidden_dim)
        hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)
        assert hidden_dim % mp_size == 0

        self.dim = dim
        self.hidden_dim = hidden_dim

        self.w1 = nn.Linear(
            dim,
            hidden_dim,
            bias=False,
        )
        self.w3 = nn.Linear(
            dim,
            hidden_dim,
            bias=False,
        )
        self.w2 = nn.Linear(
            hidden_dim,
            dim,
            bias=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # B S D
        x1 = self.w1(x.view_as(x))
        x3 = self.w3(x.view_as(x))
        output = self.w2(F.silu(x1) * x3)
        return output

    def reset_parameters(self, init_std=None, factor=1.0):
        in_init_std = init_std or (self.dim ** (-0.5))
        out_init_std = init_std or (self.hidden_dim ** (-0.5))
        in_init_std = in_init_std
        out_init_std = out_init_std / factor
        for w in [self.w1, self.w3]:
            nn.init.trunc_normal_(
                w.weight,
                mean=0.0,
                std=in_init_std,
                a=-3 * in_init_std,
                b=3 * in_init_std,
            )
        nn.init.trunc_normal_(
            self.w2.weight,
            mean=0.0,
            std=out_init_std,
            a=-3 * out_init_std,
            b=3 * out_init_std,
        )


class TransformerBlock(nn.Module):
    def __init__(self, args: BaseTransformerArgs):
        super().__init__()

        assert (args.head_dim is not None) or (
            args.n_heads is not None
        ), "Should specify at least head_dim or n_heads"
        self.head_dim = args.head_dim or args.dim // args.n_heads
        self.n_heads = args.n_heads or args.dim // args.head_dim
        self.n_kv_heads = args.n_kv_heads or self.n_heads

        assert args.n_heads % self.n_kv_heads == 0
        assert args.dim % args.n_heads == 0

        self.attention = Attention(
            dim=args.dim,
            head_dim=self.head_dim,
            n_heads=self.n_heads,
            n_kv_heads=self.n_kv_heads,
            rope_theta=args.rope_theta,
        )
        self.feed_forward = FeedForward(
            dim=args.dim,
            hidden_dim=4 * args.dim,
            multiple_of=args.multiple_of,
            ffn_dim_multiplier=args.ffn_dim_multiplier,
        )
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)

    def forward(
        self,
        x: torch.Tensor,
        freq_cis: torch.Tensor,
        tok_idx: Optional[torch.Tensor] = None,
        mask: Optional[Union[BlockMask, AttentionBias, str]] = None,
        attn_impl: str = "sdpa",
    ) -> torch.Tensor:

        h = x + self.attention(
            self.attention_norm(x),
            freq_cis,
            tok_idx=tok_idx,
            mask=mask,
            attn_impl=attn_impl,
        )
        out = h + self.feed_forward(self.ffn_norm(h))
        return out

    def init_weights(self, init_std=None, factor=1.0):
        self.attention.reset_parameters(init_std, factor)
        self.attention_norm.reset_parameters()

        self.feed_forward.reset_parameters(init_std, factor)
        self.ffn_norm.reset_parameters()


class BaseTransformer(nn.Module):
    def __init__(self, args: BaseTransformerArgs):
        super().__init__()
        self.dim = args.dim
        self.init_base_std = args.init_base_std
        self.init_std_factor = InitStdFactor(args.init_std_factor)
        self.max_seqlen = args.max_seqlen
        self.rope_embeddings = RotaryEmbedding(
            theta=args.rope_theta,
            head_dim=args.head_dim or args.dim // args.n_heads,
            max_seqlen=args.max_seqlen,
            scale_factor=args.rope_scale_factor,
            low_freq_factor=args.low_freq_factor,
            high_freq_factor=args.high_freq_factor,
            old_context_len=args.old_context_len,
        )

        self.layers = nn.ModuleList()
        for _ in range(args.n_layers):
            self.layers.append(TransformerBlock(args))

    def forward(
        self,
        h,
        tok_idx: Optional[torch.Tensor] = None,
        mask: Optional[Union[BlockMask, AttentionBias, str]] = None,
        attn_impl: str = "sdpa",
    ):

        freq_cis = self.rope_embeddings(seqlen=self.max_seqlen, tok_idx=tok_idx)

        for i, layer in enumerate(self.layers):
            h = layer(h, freq_cis, tok_idx=tok_idx, mask=mask, attn_impl=attn_impl)
        return h

    def reset_parameters(self):
        # Either use fixed base std or sqrt model dim
        self.rope_embeddings.reset_parameters()

    def init_weights(self):
        self.reset_parameters()
        for depth, layer in enumerate(self.layers):
            factor = {
                InitStdFactor.CURRENT_DEPTH: (2 * (depth + 1)) ** 0.5,
                InitStdFactor.GLOBAL_DEPTH: (2 * (len(self.layers) + 1)) ** 0.5,
                InitStdFactor.DIM_RATIO: self.dim / 4096,
                InitStdFactor.DISABLED: 1.0,
            }[self.init_std_factor]

            layer.init_weights(self.init_base_std, factor)



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

from .base_processor import BaseProcessor

from .blip_processors import (
    BlipImageTrainProcessor,
    Blip2ImageTrainProcessor,
    BlipImageEvalProcessor,
    BlipCaptionProcessor,
)

from ..common.registry import registry

__all__ = [
    "BaseProcessor",
    # BLIP
    "BlipImageTrainProcessor",
    "Blip2ImageTrainProcessor",
    "BlipImageEvalProcessor",
    "BlipCaptionProcessor",
]


def load_processor(name, cfg=None):
    """
    Example

    >>> processor = load_processor("alpro_video_train", cfg=None)
    """
    processor = registry.get_processor_class(name).from_config(cfg)

    return processor



"""VideoScore multi-dimensional video quality module.

5 dimensions: visual quality, temporal consistency, dynamic degree,
text-video alignment, factual consistency. EMNLP 2024.
"""

import logging
from typing import Optional

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class VideoScoreModule(PipelineModule):
    name = "videoscore"
    description = "VideoScore 5-dimensional video quality assessment (1-4 scale)"
    default_config = {
        "model_name": "TIGER-Lab/VideoScore",
        "num_frames": 8,
        "trust_remote_code": True,
        "model_revision": None,
    }

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._model = None
        self._processor = None

    def setup(self) -> None:
        try:
            import torch
            from transformers import AutoModelForCausalLM, AutoProcessor

            model_name = self.config.get("model_name", "TIGER-Lab/VideoScore")
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

            trc = self.config.get("trust_remote_code", True)
            rev = self.config.get("model_revision", None)
            self._processor = AutoProcessor.from_pretrained(model_name, trust_remote_code=trc, revision=rev)
            self._model = AutoModelForCausalLM.from_pretrained(
                model_name,
                torch_dtype=torch.float16 if torch.cuda.is_available() else torch.float32,
                trust_remote_code=trc,
                revision=rev,
            ).to(device)
            self._model.eval()
            self._device = device
            self._ml_available = True
            logger.info("VideoScore model loaded on %s", device)
        except (ImportError, Exception) as e:
            logger.warning("VideoScore unavailable: %s", e)

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not self._ml_available:
            return sample

        try:
            scores = self._compute_scores(sample)
            if scores:
                sample.quality_metrics.videoscore_visual = scores.get("visual_quality")
                sample.quality_metrics.videoscore_temporal = scores.get("temporal_consistency")
                sample.quality_metrics.videoscore_dynamic = scores.get("dynamic_degree")
                sample.quality_metrics.videoscore_alignment = scores.get("text_video_alignment")
                sample.quality_metrics.videoscore_factual = scores.get("factual_consistency")
        except Exception as e:
            logger.warning("VideoScore processing failed: %s", e)
        return sample

    def _compute_scores(self, sample: Sample) -> Optional[dict]:
        import cv2
        import torch
        from PIL import Image

        num_frames = self.config.get("num_frames", 8)

        # Load frames
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    frames.append(Image.fromarray(rgb))
            cap.release()
        else:
            img = Image.open(str(sample.path)).convert("RGB")
            frames = [img]

        if not frames:
            return None

        caption = sample.caption.text if sample.caption else "a video"
        dimensions = [
            "visual_quality",
            "temporal_consistency",
            "dynamic_degree",
            "text_video_alignment",
            "factual_consistency",
        ]

        scores = {}
        for dim in dimensions:
            prompt = self._build_prompt(dim, caption)
            try:
                inputs = self._processor(
                    text=prompt, images=frames, return_tensors="pt"
                ).to(self._device)

                with torch.no_grad():
                    outputs = self._model.generate(
                        **inputs, max_new_tokens=16, do_sample=False
                    )

                response = self._processor.decode(outputs[0], skip_special_tokens=True)
                score = self._extract_score(response)
                if score is not None:
                    scores[dim] = score
            except Exception as e:
                logger.debug("VideoScore dim %s failed: %s", dim, e)

        return scores if scores else None

    def _build_prompt(self, dimension: str, caption: str) -> str:
        dim_prompts = {
            "visual_quality": "Rate the visual quality (clarity, resolution, color) of this video on a scale of 1-4.",
            "temporal_consistency": "Rate the temporal consistency (smoothness, coherence) of this video on a scale of 1-4.",
            "dynamic_degree": "Rate the dynamic degree (amount of meaningful motion) of this video on a scale of 1-4.",
            "text_video_alignment": f"Rate how well this video matches the description '{caption}' on a scale of 1-4.",
            "factual_consistency": "Rate the factual consistency (adherence to real-world common sense) of this video on a scale of 1-4.",
        }
        return dim_prompts.get(dimension, "Rate the quality of this video on a scale of 1-4.")

    def _extract_score(self, response: str) -> Optional[float]:
        """Extract numeric score from model response."""
        import re

        numbers = re.findall(r"(\d+\.?\d*)", response.split(":")[-1] if ":" in response else response)
        for num_str in numbers:
            val = float(num_str)
            if 0.0 <= val <= 5.0:
                return val
        return None



# from MMF: https://github.com/facebookresearch/mmf/blob/master/mmf/utils/logger.py
# Copyright (c) Facebook, Inc. and its affiliates.

import functools
import logging
import os
import sys
import time
import wandb
from typing import Any, Dict, Union

import torch
from .distributed import get_rank, is_main_process
from termcolor import colored


def log_dict_to_wandb(log_dict, step, prefix=""):
    """include a separator `/` at the end of `prefix`"""
    if not is_main_process():
        return

    log_dict = {f"{prefix}{k}": v for k, v in log_dict.items()}
    wandb.log(log_dict, step)


def setup_wandb(config):
    if not (config.wandb.enable and is_main_process()):
        return

    run = wandb.init(
        config=config,
        project=config.wandb.project,
        entity=config.wandb.entity,
        name=os.path.basename(config.output_dir),
        reinit=True
    )
    return run


def setup_output_folder(save_dir: str, folder_only: bool = False):
    """Sets up and returns the output file where the logs will be placed
    based on the configuration passed. Usually "save_dir/logs/log_<timestamp>.txt".
    If env.log_dir is passed, logs will be directly saved in this folder.
    Args:
        folder_only (bool, optional): If folder should be returned and not the file.
            Defaults to False.
    Returns:
        str: folder or file path depending on folder_only flag
    """
    log_filename = "train_"
    log_filename += time.strftime("%Y_%m_%dT%H_%M_%S")
    log_filename += ".log"

    log_folder = os.path.join(save_dir, "logs")

    if not os.path.exists(log_folder):
        os.path.mkdirs(log_folder)

    if folder_only:
        return log_folder

    log_filename = os.path.join(log_folder, log_filename)

    return log_filename


def setup_logger(
    output: str = None,
    color: bool = True,
    name: str = "mmf",
    disable: bool = False,
    clear_handlers=True,
    *args,
    **kwargs,
):
    """
    Initialize the MMF logger and set its verbosity level to "INFO".
    Outside libraries shouldn't call this in case they have set there
    own logging handlers and setup. If they do, and don't want to
    clear handlers, pass clear_handlers options.
    The initial version of this function was taken from D2 and adapted
    for MMF.
    Args:
        output (str): a file name or a directory to save log.
            If ends with ".txt" or ".log", assumed to be a file name.
            Default: Saved to file <save_dir/logs/log_[timestamp].txt>
        color (bool): If false, won't log colored logs. Default: true
        name (str): the root module name of this logger. Defaults to "mmf".
        disable: do not use
        clear_handlers (bool): If false, won't clear existing handlers.
    Returns:
        logging.Logger: a logger
    """
    if disable:
        return None
    logger = logging.getLogger(name)
    logger.propagate = False

    logging.captureWarnings(True)
    warnings_logger = logging.getLogger("py.warnings")

    plain_formatter = logging.Formatter(
        "%(asctime)s | %(levelname)s | %(name)s : %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )

    distributed_rank = get_rank()
    handlers = []

    logging_level = logging.INFO
    # logging_level = logging.DEBUG

    if distributed_rank == 0:
        logger.setLevel(logging_level)
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(logging_level)
        if color:
            formatter = ColorfulFormatter(
                colored("%(asctime)s | %(name)s: ", "green") + "%(message)s",
                datefmt="%Y-%m-%dT%H:%M:%S",
            )
        else:
            formatter = plain_formatter
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        warnings_logger.addHandler(ch)
        handlers.append(ch)

    # file logging: all workers
    if output is None:
        output = setup_output_folder()

    if output is not None:
        if output.endswith(".txt") or output.endswith(".log"):
            filename = output
        else:
            filename = os.path.join(output, "train.log")
        if distributed_rank > 0:
            filename = filename + f".rank{distributed_rank}"
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        fh = logging.StreamHandler(_cached_log_stream(filename))
        fh.setLevel(logging_level)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)
        warnings_logger.addHandler(fh)
        handlers.append(fh)

        # Slurm/FB output, only log the main process
        #     save_dir = get_mmf_env(key="save_dir")
        if "train.log" not in filename and distributed_rank == 0:
            filename = os.path.join(output, "train.log")
            sh = logging.StreamHandler(_cached_log_stream(filename))
            sh.setLevel(logging_level)
            sh.setFormatter(plain_formatter)
            logger.addHandler(sh)
            warnings_logger.addHandler(sh)
            handlers.append(sh)

        # logger.info(f"Logging to: {filename}")

    # Remove existing handlers to add MMF specific handlers
    if clear_handlers:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
    # Now, add our handlers.
    logging.basicConfig(level=logging_level, handlers=handlers)

    return logger


def setup_very_basic_config(color=True):
    plain_formatter = logging.Formatter(
        "%(asctime)s | %(levelname)s | %(name)s : %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.INFO)
    if color:
        formatter = ColorfulFormatter(
            colored("%(asctime)s | %(name)s: ", "green") + "%(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S",
        )
    else:
        formatter = plain_formatter
    ch.setFormatter(formatter)
    # Setup a minimal configuration for logging in case something tries to
    # log a message even before logging is setup by MMF.
    logging.basicConfig(level=logging.INFO, handlers=[ch])


# cache the opened file object, so that different calls to `setup_logger`
# with the same file name can safely write to the same file.
@functools.lru_cache(maxsize=None)
def _cached_log_stream(filename):
    return open(filename, "a")


# ColorfulFormatter is adopted from Detectron2 and adapted for MMF
class ColorfulFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def formatMessage(self, record):
        log = super().formatMessage(record)
        if record.levelno == logging.WARNING:
            prefix = colored("WARNING", "red", attrs=["blink"])
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            prefix = colored("ERROR", "red", attrs=["blink", "underline"])
        else:
            return log
        return prefix + " " + log


class TensorboardLogger:
    def __init__(self, log_folder="./logs", iteration=0):
        # This would handle warning of missing tensorboard
        from torch.utils.tensorboard import SummaryWriter

        self.summary_writer = None
        self._is_master = is_main_process()
        # self.timer = Timer()
        self.log_folder = log_folder

        if self._is_master:
            # current_time = self.timer.get_time_hhmmss(None, format=self.time_format)
            current_time = time.strftime("%Y-%m-%dT%H:%M:%S")
            # self.timer.get_time_hhmmss(None, format=self.time_format)
            tensorboard_folder = os.path.join(
                self.log_folder, f"tensorboard_{current_time}"
            )
            self.summary_writer = SummaryWriter(tensorboard_folder)

    def __del__(self):
        if getattr(self, "summary_writer", None) is not None:
            self.summary_writer.close()

    def _should_log_tensorboard(self):
        if self.summary_writer is None or not self._is_master:
            return False
        else:
            return True

    def add_scalar(self, key, value, iteration):
        if not self._should_log_tensorboard():
            return

        self.summary_writer.add_scalar(key, value, iteration)

    def add_scalars(self, scalar_dict, iteration):
        if not self._should_log_tensorboard():
            return

        for key, val in scalar_dict.items():
            self.summary_writer.add_scalar(key, val, iteration)

    def add_histogram_for_model(self, model, iteration):
        if not self._should_log_tensorboard():
            return

        for name, param in model.named_parameters():
            np_param = param.clone().cpu().data.numpy()
            self.summary_writer.add_histogram(name, np_param, iteration)



"""BVQI — Blind Video Quality Index.

ICME 2023 Oral → TIP — zero-shot VQA that outperforms supervised
methods. Uses CLIP features with quality anchor texts for zero-shot
quality prediction.

GitHub: https://github.com/VQAssessment/BVQI

bvqi_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class BVQIModule(PipelineModule):
    name = "bvqi"
    description = "BVQI zero-shot blind video quality index (ICME 2023)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import bvqi
            self._model = bvqi
            self._backend = "native"
            logger.info("BVQI (native) initialised")
            return
        except ImportError:
            pass

        try:
            import pyiqa
            self._model = pyiqa.create_metric("bvqi", device="cpu")
            self._backend = "pyiqa"
            logger.info("BVQI (pyiqa) initialised")
            return
        except (ImportError, Exception):
            pass

        self._backend = "heuristic"
        logger.info("BVQI (heuristic) — install bvqi or pyiqa for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            if self._backend == "native":
                score = float(self._model.predict(str(sample.path)))
            elif self._backend == "pyiqa":
                score = self._process_pyiqa(sample)
            else:
                score = self._process_heuristic(sample)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.bvqi_score = score

        except Exception as e:
            logger.warning(f"BVQI failed for {sample.path}: {e}")

        return sample

    def _process_pyiqa(self, sample: Sample) -> Optional[float]:
        import torch
        with torch.no_grad():
            result = self._model(str(sample.path))
        return float(result.item()) if hasattr(result, "item") else float(result)

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: zero-shot quality via spatial+temporal NSS features."""
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return None
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)

        if not frames:
            return None

        spatial_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            sharpness = min(cv2.Laplacian(gray, cv2.CV_64F).var() / 600.0, 1.0)
            contrast = min(gray.std() / 65.0, 1.0)
            noise_est = cv2.Sobel(gray, cv2.CV_64F, 1, 1).var()
            cleanliness = 1.0 / (1.0 + noise_est * 0.0001)
            spatial_scores.append(0.4 * sharpness + 0.3 * contrast + 0.3 * cleanliness)

        spatial = float(np.mean(spatial_scores))

        # Temporal consistency
        if len(frames) > 1:
            diffs = []
            for i in range(len(frames) - 1):
                g1 = cv2.cvtColor(frames[i], cv2.COLOR_BGR2GRAY).astype(float)
                g2 = cv2.cvtColor(frames[i + 1], cv2.COLOR_BGR2GRAY).astype(float)
                diffs.append(np.mean(np.abs(g1 - g2)))
            temporal = 1.0 / (1.0 + np.var(diffs) * 0.01)
        else:
            temporal = 1.0

        return float(np.clip(0.7 * spatial + 0.3 * temporal, 0.0, 1.0))



"""Tests for motion_amplitude module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_motion_amplitude_basics():
    from ayase.modules.motion_amplitude import MotionAmplitudeModule
    _test_module_basics(MotionAmplitudeModule, "motion_amplitude")

def test_motion_amplitude_video(video_sample):
    from ayase.modules.motion_amplitude import MotionAmplitudeModule
    video_sample.quality_metrics = QualityMetrics()
    m = MotionAmplitudeModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for psnr99 module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_psnr99_basics():
    from ayase.modules.psnr99 import PSNR99Module
    _test_module_basics(PSNR99Module, "psnr99")

def test_psnr99_no_reference(image_sample):
    from ayase.modules.psnr99 import PSNR99Module
    m = PSNR99Module()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for stream_metric module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_stream_metric_basics():
    from ayase.modules.stream_metric import STREAMModule
    _test_module_basics(STREAMModule, "stream_metric")

def test_stream_metric_extract(video_sample):
    from ayase.modules.stream_metric import STREAMModule
    m = STREAMModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



from ayase.models import QualityMetrics


def test_tonal_dynamic_range_basics():
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule
    from .conftest import _test_module_basics

    _test_module_basics(TonalDynamicRangeModule, "tonal_dynamic_range")


def test_tonal_dynamic_range_config():
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule

    m = TonalDynamicRangeModule()
    assert "low_percentile" in m.default_config
    assert "high_percentile" in m.default_config
    assert "subsample" in m.default_config


def test_tonal_dynamic_range_image(image_sample):
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule

    m = TonalDynamicRangeModule()
    result = m.process(image_sample)
    assert result.quality_metrics is not None
    score = result.quality_metrics.tonal_dynamic_range
    assert score is not None
    assert 0.0 <= score <= 100.0


def test_tonal_dynamic_range_video(video_sample):
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule

    m = TonalDynamicRangeModule()
    result = m.process(video_sample)
    assert result.quality_metrics is not None
    score = result.quality_metrics.tonal_dynamic_range
    assert score is not None
    assert 0.0 <= score <= 100.0


def test_tonal_dynamic_range_custom_percentiles(image_sample):
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule

    m_wide = TonalDynamicRangeModule({"low_percentile": 1, "high_percentile": 99})
    m_narrow = TonalDynamicRangeModule({"low_percentile": 10, "high_percentile": 90})
    r_wide = m_wide.process(image_sample)
    # Reset metrics for second pass
    image_sample.quality_metrics = QualityMetrics()
    r_narrow = m_narrow.process(image_sample)
    # Wider percentile range should give >= score than narrower
    assert r_wide.quality_metrics.tonal_dynamic_range >= r_narrow.quality_metrics.tonal_dynamic_range



import numpy as np
import torch
import logging

logger = logging.getLogger(__name__)

# --------------------------------------------------------
# 3D sine-cosine position embedding
# References:
# MVD: https://github.com/ruiwang2021/mvd/blob/main/modeling_finetune.py
# --------------------------------------------------------
def get_3d_sincos_pos_embed(embed_dim, grid_size, t_size, cls_token=False):
    """
    grid_size: int of the grid height and width
    t_size: int of the temporal size
    return:
    pos_embed: [t_size*grid_size*grid_size, embed_dim] or [1+t_size*grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    assert embed_dim % 4 == 0
    embed_dim_spatial = embed_dim // 4 * 3
    embed_dim_temporal = embed_dim // 4

    # spatial
    grid_h = np.arange(grid_size, dtype=np.float32)
    grid_w = np.arange(grid_size, dtype=np.float32)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size, grid_size])
    pos_embed_spatial = get_2d_sincos_pos_embed_from_grid(
        embed_dim_spatial, grid
    )

    # temporal
    grid_t = np.arange(t_size, dtype=np.float32)
    pos_embed_temporal = get_1d_sincos_pos_embed_from_grid(
        embed_dim_temporal, grid_t
    )

    # concate: [T, H, W] order
    pos_embed_temporal = pos_embed_temporal[:, np.newaxis, :]
    pos_embed_temporal = np.repeat(
        pos_embed_temporal, grid_size**2, axis=1
    )  # [T, H*W, D // 4]
    pos_embed_spatial = pos_embed_spatial[np.newaxis, :, :]
    pos_embed_spatial = np.repeat(
        pos_embed_spatial, t_size, axis=0
    )  # [T, H*W, D // 4 * 3]

    pos_embed = np.concatenate([pos_embed_temporal, pos_embed_spatial], axis=-1)
    pos_embed = pos_embed.reshape([-1, embed_dim])  # [T*H*W, D]

    if cls_token:
        pos_embed = np.concatenate(
            [np.zeros([1, embed_dim]), pos_embed], axis=0
        )
    return pos_embed


# --------------------------------------------------------
# 2D sine-cosine position embedding
# References:
# Transformer: https://github.com/tensorflow/models/blob/master/official/nlp/transformer/model_utils.py
# MoCo v3: https://github.com/facebookresearch/moco-v3
# --------------------------------------------------------
def get_2d_sincos_pos_embed(embed_dim, grid_size, cls_token=False):
    """
    grid_size: int of the grid height and width
    return:
    pos_embed: [grid_size*grid_size, embed_dim] or [1+grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    grid_h = np.arange(grid_size, dtype=np.float32)
    grid_w = np.arange(grid_size, dtype=np.float32)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size, grid_size])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token:
        pos_embed = np.concatenate(
            [np.zeros([1, embed_dim]), pos_embed], axis=0
        )
    return pos_embed


def get_1d_sincos_pos_embed(embed_dim, t_size, cls_token=False):
    """
    t_size: int of the temporal size
    return:
    pos_embed: [t_size, embed_dim] or [1+t_size, embed_dim] (w/ or w/o cls_token)
    """
    grid_t = np.arange(t_size, dtype=np.float32)
    pos_embed = get_1d_sincos_pos_embed_from_grid(embed_dim, grid_t)
    if cls_token:
        pos_embed = np.concatenate(
            [np.zeros([1, embed_dim]), pos_embed], axis=0
        )
    return pos_embed


def get_2d_sincos_pos_embed_from_grid(embed_dim, grid):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(
        embed_dim // 2, grid[0]
    )  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(
        embed_dim // 2, grid[1]
    )  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1)  # (H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float32)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb


def interpolate_pos_embed(checkpoint_model, model, orig_t_size=4, pos_name='vision_encoder.pos_embed'):
    if pos_name in checkpoint_model:
        pos_embed_checkpoint = checkpoint_model[pos_name]
        embedding_size = pos_embed_checkpoint.shape[-1] # channel dim
        num_patches = model.patch_embed.num_patches # 
        num_extra_tokens = model.pos_embed.shape[-2] - num_patches # 0/1

        # we use 4 frames for pretraining
        new_t_size = model.T
        # height (== width) for the checkpoint position embedding
        orig_size = int(((pos_embed_checkpoint.shape[-2] - num_extra_tokens)//(orig_t_size)) ** 0.5)
        # height (== width) for the new position embedding
        new_size = int((num_patches // (new_t_size))** 0.5)
        
        # class_token and dist_token are kept unchanged
        if orig_t_size != new_t_size:
            logger.info(f"Temporal interpolate from {orig_t_size} to {new_t_size} ({pos_name})")
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            # B, L, C -> B， T, HW, C -> BHW, C, T  (B = 1)
            pos_tokens = pos_tokens.view(1, orig_t_size, -1, embedding_size)
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, embedding_size, orig_t_size)
            pos_tokens = torch.nn.functional.interpolate(pos_tokens, size=new_t_size, mode='linear')
            pos_tokens = pos_tokens.view(1, -1, embedding_size, new_t_size)
            pos_tokens = pos_tokens.permute(0, 3, 1, 2).reshape(1, -1, embedding_size)
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model[pos_name] = new_pos_embed
            pos_embed_checkpoint = new_pos_embed

        # class_token and dist_token are kept unchanged
        if orig_size != new_size:
            logger.info(f"Position interpolate from {orig_size}x{orig_size} to {new_size}x{new_size} ({pos_name})")
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            # B, L, C -> BT, H, W, C -> BT, C, H, W
            pos_tokens = pos_tokens.reshape(-1, new_t_size, orig_size, orig_size, embedding_size)
            pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
            pos_tokens = torch.nn.functional.interpolate(
                pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
            # BT, C, H, W -> BT, H, W, C ->  B, T, H, W, C
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, new_t_size, new_size, new_size, embedding_size) 
            pos_tokens = pos_tokens.flatten(1, 3) # B, L, C
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model[pos_name] = new_pos_embed


def interpolate_pos_embed_internvideo2(checkpoint_model, model, orig_t_size = 8):
    # interpolate position embedding
    for pos_name in ['pos_embed', 'clip_pos_embed']:
        if pos_name in checkpoint_model:
            pos_embed_checkpoint = checkpoint_model[pos_name]
            embedding_size = pos_embed_checkpoint.shape[-1] # channel dim
            num_patches = model.patch_embed.num_patches # 
            num_extra_tokens = model.pos_embed.shape[-2] - num_patches # 0/1

            # we use 8 frames for pretraining
            # new_t_size = args.num_frames * args.num_segments // model.patch_embed.tubelet_size
            new_t_size = model.num_frames // model.tubelet_size
            # height (== width) for the checkpoint position embedding
            orig_size = int(((pos_embed_checkpoint.shape[-2] - num_extra_tokens)//(orig_t_size)) ** 0.5)
            # height (== width) for the new position embedding
            new_size = int((num_patches // (new_t_size))** 0.5)
            
            # class_token and dist_token are kept unchanged
            if orig_t_size != new_t_size:
                logger.info(f"Temporal interpolate from {orig_t_size} to {new_t_size} ({pos_name})")
                extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
                # only the position tokens are interpolated
                pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
                # B, L, C -> B， T, HW, C -> BHW, C, T  (B = 1)
                pos_tokens = pos_tokens.view(1, orig_t_size, -1, embedding_size)
                pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, embedding_size, orig_t_size)
                pos_tokens = torch.nn.functional.interpolate(pos_tokens, size=new_t_size, mode='linear')
                pos_tokens = pos_tokens.view(1, -1, embedding_size, new_t_size)
                pos_tokens = pos_tokens.permute(0, 3, 1, 2).reshape(1, -1, embedding_size)
                new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
                checkpoint_model[pos_name] = new_pos_embed
                pos_embed_checkpoint = new_pos_embed

            # class_token and dist_token are kept unchanged
            if orig_size != new_size:
                logger.info(f"Position interpolate from {orig_size}x{orig_size} to {new_size}x{new_size} ({pos_name})")
                extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
                # only the position tokens are interpolated
                pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
                # B, L, C -> BT, H, W, C -> BT, C, H, W
                pos_tokens = pos_tokens.reshape(-1, new_t_size, orig_size, orig_size, embedding_size)
                pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
                pos_tokens = torch.nn.functional.interpolate(
                    pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
                # BT, C, H, W -> BT, H, W, C ->  B, T, H, W, C
                pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, new_t_size, new_size, new_size, embedding_size) 
                pos_tokens = pos_tokens.flatten(1, 3) # B, L, C
                new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
                checkpoint_model[pos_name] = new_pos_embed
    
    if 'pos_embed_spatial' in checkpoint_model or 'pos_embed_temporal' in checkpoint_model:
        raise NotImplementedError


def interpolate_pos_embed_internvideo2_new(checkpoint_model, model, orig_t_size = 8):
    pos_names = []
    for k in checkpoint_model.keys():
        if ('pos_embed' in k or 'clip_pos_embed' in k) and 'img_pos_embed' not in k:
            pos_names.append(k)
    
    logger.info(f"pos names list for interpolating: {pos_names}")

    assert len(pos_names) > 0, checkpoint_model.keys()

    if 'pos_embed_spatial' in checkpoint_model.keys() or 'pos_embed_temporal' in checkpoint_model.keys():
        raise NotImplementedError
    
    # interpolate position embedding
    for pos_name in pos_names:

        pos_embed_checkpoint = checkpoint_model[pos_name]
        embedding_size = pos_embed_checkpoint.shape[-1] # channel dim
        num_patches = model.patch_embed.num_patches # 
        num_extra_tokens = model.pos_embed.shape[-2] - num_patches # 0/1

        # we use 8 frames for pretraining
        # new_t_size = args.num_frames * args.num_segments // model.patch_embed.tubelet_size
        new_t_size = model.num_frames // model.tubelet_size
        # height (== width) for the checkpoint position embedding
        orig_size = int(((pos_embed_checkpoint.shape[-2] - num_extra_tokens)//(orig_t_size)) ** 0.5)
        # height (== width) for the new position embedding
        new_size = int((num_patches // (new_t_size))** 0.5)
        
        # class_token and dist_token are kept unchanged
        if orig_t_size != new_t_size:
            logger.info(f"Temporal interpolate from {orig_t_size} to {new_t_size} ({pos_name})")
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            # B, L, C -> B， T, HW, C -> BHW, C, T  (B = 1)
            pos_tokens = pos_tokens.view(1, orig_t_size, -1, embedding_size)
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, embedding_size, orig_t_size)
            pos_tokens = torch.nn.functional.interpolate(pos_tokens, size=new_t_size, mode='linear')
            pos_tokens = pos_tokens.view(1, -1, embedding_size, new_t_size)
            pos_tokens = pos_tokens.permute(0, 3, 1, 2).reshape(1, -1, embedding_size)
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model[pos_name] = new_pos_embed
            pos_embed_checkpoint = new_pos_embed

        # class_token and dist_token are kept unchanged
        if orig_size != new_size:
            logger.info(f"Position interpolate from {orig_size}x{orig_size} to {new_size}x{new_size} ({pos_name})")
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            # B, L, C -> BT, H, W, C -> BT, C, H, W
            pos_tokens = pos_tokens.reshape(-1, new_t_size, orig_size, orig_size, embedding_size)
            pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
            pos_tokens = torch.nn.functional.interpolate(
                pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
            # BT, C, H, W -> BT, H, W, C ->  B, T, H, W, C
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).reshape(-1, new_t_size, new_size, new_size, embedding_size) 
            pos_tokens = pos_tokens.flatten(1, 3) # B, L, C
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model[pos_name] = new_pos_embed


"""Tests for vmaf module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_vmaf_basics():
    from ayase.modules.vmaf import VMAFModule
    _test_module_basics(VMAFModule, "vmaf")

def test_vmaf_no_reference(image_sample):
    from ayase.modules.vmaf import VMAFModule
    m = VMAFModule()
    result = m.process(image_sample)
    assert result is image_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.

import json
import os
import shutil
import subprocess
from dataclasses import dataclass
from typing import Any, Dict

from omegaconf import OmegaConf

from ..core.args import dataclass_from_dict


@dataclass
class StoolArgs:
    config: Any = None
    launcher: str = "sbatch"  # Can be sbatch or bash if already in salloc
    script: str = "apps.main.train"  # The script to run.
    copy_code: bool = True  # Wether to copy code to dump dir
    dirs_exists_ok: bool = (
        False  # Wether to copy new code and config and run regardless that dir exists
    )
    override: bool = False  # Wether to delete dump dir and restart
    nodes: int = -1  # The number of nodes to run the job on.
    ngpu: int = 8  # The number of GPUs required per node.
    ncpu: int = 16  # The number of CPUs allocated per GPU.
    mem: str = ""  # The amount of memory to allocate.
    anaconda: str = "default"  # The path to the anaconda environment.
    constraint: str = ""  # The constraint on the nodes.
    exclude: str = ""  # The nodes to exclude.
    time: int = -1  # The time limit of the job (in minutes).
    account: str = ""
    qos: str = ""
    partition: str = "learn"
    stdout: bool = False


SBATCH_COMMAND = """#!/bin/bash

{exclude}
{qos}
{account}
{constraint}
#SBATCH --job-name={name}
#SBATCH --nodes={nodes}
#SBATCH --gres=gpu:{ngpus}
#SBATCH --cpus-per-gpu={ncpu}
#SBATCH --time={time}
#SBATCH --partition={partition}
#SBATCH --mem={mem}

#SBATCH --output={dump_dir}/logs/%j/%j.stdout
#SBATCH --error={dump_dir}/logs/%j/%j.stderr

#SBATCH --open-mode=append
#SBATCH --signal=USR2@120
#SBATCH --distribution=block

# Mimic the effect of "conda init", which doesn't work for scripts
eval "$({conda_exe} shell.bash hook)"
source activate {conda_env_path}

{go_to_code_dir}

export OMP_NUM_THREADS=1
export LAUNCH_WITH="SBATCH"
export DUMP_DIR={dump_dir}
srun {log_output} -n {tasks} -N {nodes_per_run} python -u -m {script} config=$DUMP_DIR/base_config.yaml
"""


def copy_dir(input_dir: str, output_dir: str) -> None:
    print(f"Copying : {input_dir}\n" f"to      : {output_dir} ...")
    assert os.path.isdir(input_dir), f"{input_dir} is not a directory"
    assert os.path.isdir(output_dir), f"{output_dir} is not a directory"
    rsync_cmd = (
        f"rsync -arm --copy-links "
        f"--include '**/' "
        f"--include '*.py' "
        f"--include '*.yaml' "
        f"--exclude='*' "
        f"{input_dir}/ {output_dir}"
    )
    print(f"Copying command: {rsync_cmd}")
    subprocess.call([rsync_cmd], shell=True)
    print("Copy done.")


def retrieve_max_time_per_partition() -> Dict[str, int]:
    # retrieve partition max times (a bit slow)

    sinfo = json.loads(subprocess.check_output("sinfo --json", shell=True))["sinfo"]
    max_times: Dict[str, int] = {}

    for info in sinfo:
        if info["partition"]["maximums"]["time"]["infinite"]:
            max_times[info["partition"]["name"]] = 14 * 24 * 60  # 14 days
        else:
            max_times[info["partition"]["name"]] = info["partition"]["maximums"][
                "time"
            ][
                "number"
            ]  # in minutes

    return max_times


def validate_args(args) -> None:
    # Set maximum time limit if not specified
    if args.time == -1:
        max_times = retrieve_max_time_per_partition()
        args.time = max_times.get(
            args.partition, 3 * 24 * 60
        )  # Default to 3 days if not found
        print(
            f"No time limit specified, using max time for partitions: {args.time} minutes"
        )

    if args.constraint:
        args.constraint = f"#SBATCH --constraint={args.constraint}"

    if args.account:
        args.account = f"#SBATCH  --account={args.account}"

    if args.qos:
        args.qos = f"#SBATCH --qos={args.qos}"

    if getattr(args, "exclude", ""):
        args.exclude = f"#SBATCH --exclude={args.exclude}"

    if hasattr(args, "anaconda") and args.anaconda:
        if args.anaconda == "default":
            args.anaconda = (
                subprocess.check_output("which python", shell=True)
                .decode("ascii")
                .strip()
            )
        else:
            args.anaconda = f"{args.anaconda}/bin/python"
        assert os.path.isfile(args.anaconda)

    args.mem = args.mem or "0"

    assert args.partition
    assert args.ngpu > 0
    assert args.ncpu > 0
    assert args.nodes > 0
    assert args.time > 0
    assert args.partition


def launch_job(args: StoolArgs):
    # Set up args default and validate them depending on the cluster or partition requested
    validate_args(args)
    dump_dir = args.config["dump_dir"]
    job_name = args.config["name"]
    print("Creating directories...")
    os.makedirs(dump_dir, exist_ok=args.dirs_exists_ok or args.override)
    if args.override:
        confirm = input(
            f"Are you sure you want to delete the directory '{dump_dir}'? This action cannot be undone. (yes/no): "
        )
        if confirm.lower() == "yes":
            shutil.rmtree(dump_dir)
            print(f"Directory '{dump_dir}' has been deleted.")
        else:
            print("Operation cancelled.")
            return
    if args.copy_code:
        os.makedirs(f"{dump_dir}/code", exist_ok=args.dirs_exists_ok)
        print("Copying code ...")
        copy_dir(os.getcwd(), f"{dump_dir}/code")

    print("Saving config file ...")
    with open(f"{dump_dir}/base_config.yaml", "w") as cfg:
        cfg.write(OmegaConf.to_yaml(args.config))

    conda_exe = os.environ.get("CONDA_EXE", "conda")
    conda_env_path = os.path.dirname(os.path.dirname(args.anaconda))
    log_output = (
        "-o $DUMP_DIR/logs/%j/%j_%t.out -e $DUMP_DIR/logs/%j/%j_%t.err"
        if not args.stdout
        else ""
    )
    sbatch = SBATCH_COMMAND.format(
        name=job_name,
        script=args.script,
        dump_dir=dump_dir,
        nodes=args.nodes,
        tasks=args.nodes * args.ngpu,
        nodes_per_run=args.nodes,
        ngpus=args.ngpu,
        ncpu=args.ncpu,
        mem=args.mem,
        qos=args.qos,
        account=args.account,
        constraint=args.constraint,
        exclude=args.exclude,
        time=args.time,
        partition=args.partition,
        conda_exe=conda_exe,
        conda_env_path=conda_env_path,
        log_output=log_output,
        go_to_code_dir=f"cd {dump_dir}/code/" if args.copy_code else "",
    )

    print("Writing sbatch command ...")
    with open(f"{dump_dir}/submit.slurm", "w") as f:
        f.write(sbatch)

    print("Submitting job ...")
    os.system(f"{args.launcher} {dump_dir}/submit.slurm")

    print("Done.")


if __name__ == "__main__":
    """
    The command line interface here uses OmegaConf https://omegaconf.readthedocs.io/en/2.3_branch/usage.html#from-command-line-arguments
    This accepts arguments as a dot list
    So if the dataclass looks like

    @dataclass
    class DummyArgs:
        name: str
        mode: LMTransformerArgs

    @dataclass
    class LMTransformerArgs:
        dim: int

    Then you can pass model.dim=32 to change values in LMTransformerArgs
    or just name=tictac for top level attributes.
    """
    args = OmegaConf.from_cli()
    args.config = OmegaConf.load(args.config)
    args = dataclass_from_dict(StoolArgs, args)
    launch_job(args)



"""Tests for HDR-ChipQA module."""

from ayase.models import QualityMetrics

from ..conftest import _test_module_basics


def test_hdr_chipqa_basics():
    from ayase.modules.hdr_chipqa import HDRChipQAModule

    _test_module_basics(HDRChipQAModule, "hdr_chipqa")


def test_hdr_chipqa_skip_without_backend(video_sample):
    from ayase.modules.hdr_chipqa import HDRChipQAModule

    video_sample.quality_metrics = QualityMetrics()
    result = HDRChipQAModule().process(video_sample)
    assert result is video_sample
    assert result.quality_metrics.hdr_chipqa_score is None


def test_hdr_chipqa_field_exists():
    qm = QualityMetrics()
    assert hasattr(qm, "hdr_chipqa_score")
    assert qm.hdr_chipqa_score is None


def test_hdr_chipqa_field_group():
    qm = QualityMetrics()
    assert qm._FIELD_GROUPS.get("hdr_chipqa_score") == "hdr"



"""Dynamics Controllability module.

Assesses if motion in video matches the dynamics implied by the text prompt.
Parses caption for motion keywords and compares to actual motion.
Range: 0-1 (higher = better controllability/alignment).

Backend tiers:
  1. **CoTracker** — Camera motion classification + object trajectory analysis
  2. **Farneback** — OpenCV dense optical flow (current approach)
  3. **Keyword** — Pure text keyword matching
"""

import logging
import re

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

# Camera motion keywords mapped to expected motion vectors
CAMERA_KEYWORDS = {
    "pan left": "pan_left", "pan right": "pan_right",
    "panning left": "pan_left", "panning right": "pan_right",
    "tilt up": "tilt_up", "tilt down": "tilt_down",
    "tilting up": "tilt_up", "tilting down": "tilt_down",
    "zoom in": "zoom_in", "zoom out": "zoom_out",
    "zooming in": "zoom_in", "zooming out": "zoom_out",
    "orbit": "orbit", "orbiting": "orbit",
    "tracking shot": "tracking", "dolly": "tracking",
    "static camera": "static", "fixed camera": "static",
    "handheld": "handheld", "shaky": "handheld",
}


class DynamicsControllabilityModule(PipelineModule):
    name = "dynamics_controllability"
    description = "Assesses motion controllability based on text-motion alignment"
    default_config = {
        "subsample": 16,
    }

    # Motion keywords and their expected motion levels (0-1)
    MOTION_KEYWORDS = {
        "static": 0.0, "still": 0.0, "stationary": 0.0, "motionless": 0.0,
        "slow": 0.2, "slowly": 0.2, "gentle": 0.2, "calm": 0.2,
        "moving": 0.5, "walk": 0.5, "walking": 0.5, "moderate": 0.5,
        "fast": 0.8, "quickly": 0.8, "rapid": 0.8, "swift": 0.8,
        "run": 0.8, "running": 0.8, "sprint": 0.9, "dash": 0.9, "race": 0.9,
        "dynamic": 0.7, "energetic": 0.7, "action": 0.8,
        "explosive": 0.9, "sudden": 0.9,
        "smooth": 0.4, "fluid": 0.4, "flowing": 0.4,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._backend = "farneback"
        self._cotracker = None
        self._device = None

    def setup(self) -> None:
        # Tier 1: CoTracker
        try:
            import torch

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self._cotracker = torch.hub.load(
                "facebookresearch/co-tracker", "cotracker2"
            ).to(device).eval()
            self._device = device
            self._backend = "cotracker"
            logger.info("DynamicsControllability loaded CoTracker on %s", device)
            return
        except Exception as e:
            logger.info("CoTracker unavailable for dynamics: %s", e)

        # Tier 2: Farneback
        try:
            cv2.calcOpticalFlowFarneback  # noqa: B018
            self._backend = "farneback"
            logger.info("DynamicsControllability using Farneback optical flow")
        except AttributeError:
            logger.info("DynamicsControllability using keyword-only mode")

    def _extract_expected_motion(self, caption: str) -> float:
        caption_lower = caption.lower()
        matched_levels = []
        for keyword, level in self.MOTION_KEYWORDS.items():
            if re.search(r'\b' + keyword + r'\b', caption_lower):
                matched_levels.append(level)
        if not matched_levels:
            return 0.5
        return float(np.mean(matched_levels))

    def _extract_camera_keywords(self, caption: str) -> list:
        caption_lower = caption.lower()
        found = []
        for phrase, motion_type in CAMERA_KEYWORDS.items():
            if phrase in caption_lower:
                found.append(motion_type)
        return found

    def _classify_camera_motion(self, flow: np.ndarray) -> str:
        """Classify global camera motion from dense optical flow field."""
        mean_dx = float(np.mean(flow[..., 0]))
        mean_dy = float(np.mean(flow[..., 1]))

        h, w = flow.shape[:2]
        cy, cx = h / 2, w / 2
        y_coords, x_coords = np.mgrid[0:h, 0:w]
        # Radial flow for zoom detection
        radial_x = flow[..., 0] - mean_dx
        radial_y = flow[..., 1] - mean_dy
        dir_x = (x_coords - cx) / (w / 2 + 1e-6)
        dir_y = (y_coords - cy) / (h / 2 + 1e-6)
        radial_component = radial_x * dir_x + radial_y * dir_y
        mean_radial = float(np.mean(radial_component))

        mag = np.sqrt(mean_dx ** 2 + mean_dy ** 2)
        if mag < 0.5 and abs(mean_radial) < 0.3:
            return "static"
        if mean_radial > 1.0:
            return "zoom_in"
        if mean_radial < -1.0:
            return "zoom_out"
        if abs(mean_dx) > abs(mean_dy) * 1.5:
            return "pan_left" if mean_dx < 0 else "pan_right"
        if abs(mean_dy) > abs(mean_dx) * 1.5:
            return "tilt_up" if mean_dy < 0 else "tilt_down"
        return "tracking"

    def _compute_cotracker(self, sample: Sample) -> tuple:
        """Returns (actual_motion_level, detected_camera_motion)."""
        import torch

        num_frames = self.config.get("subsample", 16)
        cap = cv2.VideoCapture(str(sample.path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]

        frames = []
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        cap.release()

        if len(frames) < 3:
            return None, "unknown"

        h, w = frames[0].shape[:2]
        target_h, target_w = min(h, 384), min(w, 512)
        resized = [cv2.resize(f, (target_w, target_h)) for f in frames]

        video_tensor = (
            torch.from_numpy(np.stack(resized))
            .permute(0, 3, 1, 2)
            .unsqueeze(0)
            .float()
            .to(self._device)
        )

        with torch.no_grad():
            pred_tracks, pred_visibility = self._cotracker(video_tensor)

        tracks = pred_tracks[0].cpu().numpy()  # [T, N, 2]
        visibility = pred_visibility[0].cpu().numpy()

        good = visibility.sum(axis=0) >= len(frames) * 0.5
        if good.sum() < 5:
            return None, "unknown"
        tracks = tracks[:, good, :]

        velocities = np.diff(tracks, axis=0)  # [T-1, N, 2]

        # Global motion (camera) = median velocity per frame
        global_vel = np.median(velocities, axis=1)  # [T-1, 2]

        # Classify camera motion from global velocity
        mean_gx = float(np.mean(global_vel[:, 0]))
        mean_gy = float(np.mean(global_vel[:, 1]))
        global_mag = np.sqrt(mean_gx ** 2 + mean_gy ** 2)

        if global_mag < 1.0:
            camera_motion = "static"
        elif abs(mean_gx) > abs(mean_gy) * 1.5:
            camera_motion = "pan_left" if mean_gx < 0 else "pan_right"
        elif abs(mean_gy) > abs(mean_gx) * 1.5:
            camera_motion = "tilt_up" if mean_gy < 0 else "tilt_down"
        else:
            camera_motion = "tracking"

        # Object motion = velocity after subtracting global
        object_vel = velocities - global_vel[:, np.newaxis, :]
        object_mag = np.sqrt(np.sum(object_vel ** 2, axis=-1))
        total_mag = np.sqrt(np.sum(velocities ** 2, axis=-1))

        avg_motion = float(np.mean(total_mag))
        normalized = min(avg_motion / 15.0, 1.0)

        return normalized, camera_motion

    def _compute_farneback(self, video_path) -> float:
        try:
            cap = cv2.VideoCapture(str(video_path))
            max_frames = self.config.get("max_frames", 300)
            motion_magnitudes = []
            prev_frame = None
            frame_count = 0

            while True:
                ret, frame = cap.read()
                if not ret:
                    break

                if frame_count >= max_frames:
                    break

                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

                if prev_frame is not None and frame_count % 3 == 0:
                    try:
                        flow = cv2.calcOpticalFlowFarneback(
                            prev_frame, gray, None, 0.5, 3, 15, 3, 5, 1.2, 0
                        )
                        magnitude, _ = cv2.cartToPolar(flow[..., 0], flow[..., 1])
                        motion_magnitudes.append(magnitude.mean())
                    except Exception:
                        pass

                prev_frame = gray
                frame_count += 1

            cap.release()

            if not motion_magnitudes:
                return None

            avg_motion = np.mean(motion_magnitudes)
            normalized_motion = min(avg_motion / 10.0, 1.0)
            return float(normalized_motion)

        except Exception as e:
            logger.debug(f"Farneback motion computation failed: {e}")
            return None

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample
        if sample.caption is None or not sample.caption.text:
            return sample

        try:
            caption = sample.caption.text
            expected_motion = self._extract_expected_motion(caption)
            camera_keywords = self._extract_camera_keywords(caption)

            actual_motion = None
            camera_motion_match = 1.0

            if self._backend == "cotracker":
                try:
                    actual_motion, detected_camera = self._compute_cotracker(sample)
                    # Camera motion keyword match bonus
                    if camera_keywords and detected_camera != "unknown":
                        if detected_camera in camera_keywords:
                            camera_motion_match = 1.0
                        elif "static" in camera_keywords and detected_camera == "static":
                            camera_motion_match = 1.0
                        else:
                            camera_motion_match = 0.5
                except Exception as e:
                    logger.info("CoTracker dynamics failed: %s, using Farneback", e)

            if actual_motion is None and self._backend in ("farneback", "cotracker"):
                actual_motion = self._compute_farneback(sample.path)

            if actual_motion is None:
                # Cannot compute actual motion — skip rather than storing a meaningless score
                return sample

            # Compute controllability
            error = abs(expected_motion - actual_motion)
            base_controllability = 1.0 - error

            # Blend with camera motion match if camera keywords present
            if camera_keywords:
                controllability = 0.7 * base_controllability + 0.3 * camera_motion_match
            else:
                controllability = base_controllability

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.dynamics_controllability = float(
                np.clip(controllability, 0.0, 1.0)
            )

            logger.debug(
                "Dynamics controllability for %s: %.3f (expected: %.2f, actual: %.2f)",
                sample.path.name, controllability, expected_motion, actual_motion,
            )

        except Exception as e:
            logger.warning("Dynamics controllability processing failed for %s: %s", sample.path, e)

        return sample



"""Tests for compare2score module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_compare2score_basics():
    from ayase.modules.compare2score import Compare2ScoreModule
    _test_module_basics(Compare2ScoreModule, "compare2score")

def test_compare2score_image(image_sample):
    from ayase.modules.compare2score import Compare2ScoreModule
    image_sample.quality_metrics = QualityMetrics()
    m = Compare2ScoreModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_compare2score_video(video_sample):
    from ayase.modules.compare2score import Compare2ScoreModule
    video_sample.quality_metrics = QualityMetrics()
    m = Compare2ScoreModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""UCIQE — Underwater Color Image Quality Evaluation.

Yang & Sowmya 2015 — pure CIELab-based metric for underwater images.
Combines chroma standard deviation, luminance contrast, and saturation.

GitHub: https://github.com/paulwong16/UCIQE

uciqe_score — higher = better quality

Formula: UCIQE = c1*σ_c + c2*con_l + c3*μ_s
Default weights: c1=0.4680, c2=0.2745, c3=0.2576
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _compute_uciqe(img: np.ndarray, c1: float, c2: float, c3: float) -> float:
    """Compute UCIQE score for a single BGR image.

    Uses standard CIELab [0-100] range (rescaled from OpenCV's [0-255]).
    Formula: UCIQE = c1*σ_c + c2*con_l + c3*μ_s  (Yang & Sowmya 2015)
    """
    # Convert to CIELab — OpenCV returns L in [0,255], a/b in [0,255] (128-centered)
    lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB).astype(np.float64)

    # Rescale to standard CIELab ranges: L [0,100], a [-128,127], b [-128,127]
    l_channel = lab[:, :, 0] * (100.0 / 255.0)
    a_channel = lab[:, :, 1] - 128.0
    b_channel = lab[:, :, 2] - 128.0

    # Chroma
    chroma = np.sqrt(a_channel ** 2 + b_channel ** 2)
    sigma_c = np.std(chroma)  # chroma std

    # Luminance contrast: top 1% - bottom 1%
    l_sorted = np.sort(l_channel.ravel())
    n = len(l_sorted)
    top_1 = np.mean(l_sorted[int(0.99 * n) :])
    bot_1 = np.mean(l_sorted[: max(int(0.01 * n), 1)])
    con_l = top_1 - bot_1

    # Saturation: S = C / L (paper formula)
    saturation = chroma / (l_channel + 1e-8)
    mu_s = np.mean(saturation)

    return c1 * sigma_c + c2 * con_l + c3 * mu_s


class UCIQEModule(PipelineModule):
    name = "uciqe"
    description = "UCIQE underwater color image quality evaluation (2015)"
    default_config = {
        "c1": 0.4680,
        "c2": 0.2745,
        "c3": 0.2576,
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.c1 = self.config.get("c1", 0.4680)
        self.c2 = self.config.get("c2", 0.2745)
        self.c3 = self.config.get("c3", 0.2576)
        self.subsample = self.config.get("subsample", 8)

    def process(self, sample: Sample) -> Sample:
        try:
            score = self._compute(sample)
            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.uciqe_score = score
        except Exception as e:
            logger.warning(f"UCIQE failed for {sample.path}: {e}")
        return sample

    def _compute(self, sample: Sample) -> Optional[float]:
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return None
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            scores = []
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    scores.append(_compute_uciqe(frame, self.c1, self.c2, self.c3))
            cap.release()
            return float(np.mean(scores)) if scores else None
        else:
            img = cv2.imread(str(sample.path))
            if img is None:
                return None
            return float(_compute_uciqe(img, self.c1, self.c2, self.c3))



import logging
import cv2
import numpy as np
from PIL import Image
import os
import json
from typing import Optional
from pathlib import Path
import urllib.request

from ayase.models import Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

# Original: https://github.com/christophschuhmann/improved-aesthetic-predictor/raw/main/sac+logos+ava1-l14-linearMSE.pth
AESTHETIC_MLP_URL = "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/aesthetic_scoring/sac+logos+ava1-l14-linearMSE.pth"
AESTHETIC_MLP_FILENAME = "sac+logos+ava1-l14-linearMSE.pth"


class AestheticScoringModule(PipelineModule):
    name = "aesthetic_scoring"
    description = "Calculates aesthetic score (1-10) using LAION-Aesthetics MLP"
    default_config = {"models_dir": "models"}

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._processor = None
        self._mlp = None
        self._device = "cpu"
        self._ml_available = False

        # URL for the weights (not downloading here, just reference)
        # https://github.com/christophschuhmann/improved-aesthetic-predictor/blob/main/sac+logos+ava1-l14-linearMSE.pth

    def setup(self) -> None:
        try:
            import torch
            import torch.nn as nn
            from transformers import CLIPModel, CLIPProcessor

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info(f"Loading Aesthetic Scoring models on {self._device}...")
            
            models_dir = self.config.get("models_dir", "models")

            # Load CLIP
            model_name = "openai/clip-vit-large-patch14"
            self._model = CLIPModel.from_pretrained(model_name, cache_dir=models_dir, use_safetensors=True).to(self._device)
            self._processor = CLIPProcessor.from_pretrained(model_name, cache_dir=models_dir)

            # Load MLP Head
            weight_dir = Path(models_dir) / "aesthetic_scoring"
            weight_dir.mkdir(parents=True, exist_ok=True)
            weight_path = weight_dir / AESTHETIC_MLP_FILENAME
            if not weight_path.exists():
                urllib.request.urlretrieve(AESTHETIC_MLP_URL, weight_path)
            
            # The official weight is a simple serialized torch dict or similar.
            # Usually it's just a linear layer weight (768, 1) and bias.
            
            state_dict = torch.load(str(weight_path), map_location=self._device, weights_only=True)
            
            # LAION Aesthetic V2 is an MLP: 768 -> 1024 -> 128 -> 64 -> 16 -> 1
            # Keys in state_dict: layers.0.weight, layers.0.bias, ..., layers.7.weight, layers.7.bias
            class _AestheticMLP(nn.Module):
                def __init__(self):
                    super().__init__()
                    self.layers = nn.Sequential(
                        nn.Linear(768, 1024), nn.Dropout(0.2),
                        nn.Linear(1024, 128), nn.Dropout(0.2),
                        nn.Linear(128, 64), nn.Dropout(0.1),
                        nn.Linear(64, 16),
                        nn.Linear(16, 1),
                    )
                def forward(self, x):
                    return self.layers(x)

            self._linear = _AestheticMLP()
            self._linear.load_state_dict(state_dict)
            self._linear.to(self._device)
            self._linear.eval()
            
            self._ml_available = True
            logger.info("Loaded LAION-Aesthetics V2 predictor.")

        except ImportError:
            logger.warning("Transformers/Torch not installed. Aesthetic scoring disabled.")
        except Exception as e:
            logger.error(f"Failed to load Aesthetic models: {e}")

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        try:
            from ayase.utils.sampling import FrameSampler
            frames = FrameSampler.sample_frames(sample.path, num_frames=8)
            
            if not frames:
                return sample

            import torch
            from PIL import Image

            # Batch process frames if possible to save overhead?
            # For simplicity and safety with varying sizes, we loop.
            
            scores = []
            
            for image in frames:
                # Convert BGR to RGB
                image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                pil_image = Image.fromarray(image_rgb)

                inputs = self._processor(images=pil_image, return_tensors="pt").to(self._device)

                with torch.no_grad():
                    image_features = self._model.get_image_features(**inputs)
                    image_features = image_features / image_features.norm(p=2, dim=-1, keepdim=True)

                    # Predict score
                    score_tensor = self._linear(image_features.float())
                    scores.append(score_tensor.item())

            if not scores:
                return sample

            avg_score = sum(scores) / len(scores)

            # Check if we have quality_metrics
            if sample.quality_metrics is None:
                from ayase.models import QualityMetrics

                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.aesthetic_mlp_score = float(avg_score)

            if avg_score < 4.5:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.INFO,
                        message=f"Low aesthetic score: {avg_score:.2f}",
                        details={"score": avg_score, "frame_scores": scores},
                    )
                )

        except Exception as e:
            logger.warning(f"Aesthetic scoring failed: {e}")

        return sample



"""Tests for vif module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_vif_basics():
    from ayase.modules.vif import VIFModule
    _test_module_basics(VIFModule, "vif")

def test_vif_no_reference(image_sample):
    from ayase.modules.vif import VIFModule
    m = VIFModule()
    result = m.process(image_sample)
    assert result is image_sample



import codecs
import json
import os
from subprocess import call

from PIL import Image
from torchvision.datasets import VisionDataset

SUPPORTED_LANGUAGES = [
    "ar",
    "bn",
    "cs",
    "da",
    "de",
    "el",
    "en",
    "es",
    "fa",
    "fi",
    "fil",
    "fr",
    "he",
    "hi",
    "hr",
    "hu",
    "id",
    "it",
    "ja",
    "ko",
    "mi",
    "nl",
    "no",
    "pl",
    "pt",
    "quz",
    "ro",
    "ru",
    "sv",
    "sw",
    "te",
    "th",
    "tr",
    "uk",
    "vi",
    "zh",
]

CAPTIONS_DOWNLOAD_URL = "https://google.github.io/crossmodal-3600/web-data/captions.zip"
IMAGES_DOWNLOAD_URL = (
    "https://open-images-dataset.s3.amazonaws.com/crossmodal-3600/images.tgz"
)
OUTPUT_FILENAME_TEMPLATE = "crossmodal3600_captions-{}.json"


class Crossmodal3600(VisionDataset):
    def __init__(self, root, ann_file, transform=None, target_transform=None):
        super().__init__(root, transform=transform, target_transform=target_transform)
        self.ann_file = os.path.expanduser(ann_file)
        with codecs.open(ann_file, "r", encoding="utf-8") as fp:
            data = json.load(fp)
        self.data = [
            (img_path, txt)
            for img_path, txt in zip(data["image_paths"], data["annotations"])
        ]

    def __getitem__(self, index):
        img, captions = self.data[index]

        # Image
        img = Image.open(img).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)

        # Captions
        target = [
            captions,
        ]
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


def _download_captions(out_path):
    os.makedirs(out_path, exist_ok=True)
    print("Downloading captions")
    call(f"wget {CAPTIONS_DOWNLOAD_URL} -O captions.zip", shell=True)
    call(f"unzip captions.zip -d {out_path}", shell=True)
    call("rm captions.zip", shell=True)


def _download_images(out_path):
    os.makedirs(out_path, exist_ok=True)
    print("Downloading images")
    call(f"wget {IMAGES_DOWNLOAD_URL} -O images.tgz", shell=True)
    call(f"tar -xzf images.tgz -C {out_path}", shell=True)
    call("rm images.tgz", shell=True)


def create_annotation_file(root, lang_code):
    if lang_code not in SUPPORTED_LANGUAGES:
        raise ValueError(
            f"Language code {lang_code} not supported. Supported languages are {SUPPORTED_LANGUAGES}"
        )
    data_dir = os.path.join(root, "xm3600")
    images_dir = os.path.join(data_dir, "images")
    if not os.path.exists(images_dir):
        _download_images(images_dir)
    captions_path = os.path.join(data_dir, "captions.jsonl")
    if not os.path.exists(captions_path):
        _download_captions(data_dir)
    with open(captions_path, "r", encoding="utf-8") as f:
        data = f.readlines()
    data = [json.loads(line) for line in data]

    number_of_missing_images = 0
    valid_images, valid_annotations, valid_indicies = [], [], []
    for i, data_item in enumerate(data):
        image_id = data_item["image/key"]
        image_name = f"{image_id}.jpg"
        image_path = os.path.join(images_dir, image_name)
        if not os.path.exists(image_path):
            print("Missing image file", image_name)
            number_of_missing_images += 1
            continue
        captions = data_item[lang_code]["caption"]
        txt = captions[0]

        valid_images.append(image_path)
        valid_annotations.append(txt)
        valid_indicies.append(i)

    if number_of_missing_images > 0:
        print(f"*** WARNING *** missing {number_of_missing_images} files.")

    with codecs.open(
        os.path.join(root, OUTPUT_FILENAME_TEMPLATE.format(lang_code)),
        "w",
        encoding="utf-8",
    ) as fp:
        json.dump(
            {
                "image_paths": valid_images,
                "annotations": valid_annotations,
                "indicies": valid_indicies,
            },
            fp,
            ensure_ascii=False,
        )



import os
import time
from contextlib import suppress

import numpy as np
import torch
import torch.nn.functional as F
from sklearn.metrics import balanced_accuracy_score, classification_report
from torch.utils.data import DataLoader, Dataset, Sampler
from tqdm import tqdm

from .zeroshot_classification import accuracy


def assign_learning_rate(param_group, new_lr):
    param_group["lr"] = new_lr


def _warmup_lr(base_lr, warmup_length, step):
    return base_lr * (step + 1) / warmup_length


def cosine_lr(optimizer, base_lrs, warmup_length, steps):
    if not isinstance(base_lrs, list):
        base_lrs = [base_lrs for _ in optimizer.param_groups]
    assert len(base_lrs) == len(optimizer.param_groups)

    def _lr_adjuster(step):
        for param_group, base_lr in zip(optimizer.param_groups, base_lrs):
            if step < warmup_length:
                lr = _warmup_lr(base_lr, warmup_length, step)
            else:
                e = step - warmup_length
                es = steps - warmup_length
                lr = 0.5 * (1 + np.cos(np.pi * e / es)) * base_lr
            assign_learning_rate(param_group, lr)

    return _lr_adjuster


class Featurizer(torch.nn.Module):
    def __init__(self, model, normalize=True):
        super().__init__()
        self.model = model
        self.normalize = normalize

    def forward(self, input):
        image_features = self.model.encode_image(input)
        if self.normalize:
            image_features = F.normalize(image_features, dim=-1)
        return image_features


class FeatureDataset(Dataset):
    def __init__(self, features, targets):
        self.features = features
        self.targets = targets

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        return self.features[i], self.targets[i]


def train(
    dataloader,
    input_shape,
    output_shape,
    weight_decay,
    lr,
    epochs,
    autocast,
    device,
    seed,
):
    torch.manual_seed(seed)
    model = torch.nn.Linear(input_shape, output_shape)
    devices = [x for x in range(torch.cuda.device_count())]
    model = model.cuda()
    model = torch.nn.DataParallel(model, device_ids=devices)
    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr=lr,
        weight_decay=weight_decay,
    )
    criterion = torch.nn.CrossEntropyLoss()

    len_loader = len(dataloader)
    scheduler = cosine_lr(optimizer, lr, 0.0, epochs * len_loader)

    for epoch in range(epochs):
        end = time.time()
        for i, (x, y) in enumerate(dataloader):
            x, y = x.cuda(), y.cuda()
            step = i + epoch * len_loader
            data_time = time.time() - end
            scheduler(step)

            optimizer.zero_grad()
            with autocast():
                pred = model(x)
                loss = criterion(pred, y)

            loss.backward()
            optimizer.step()

            batch_time = time.time() - end
            end = time.time()

            if (i % 20) == 1:
                num_samples = i * len(x)
                try:
                    samples_per_epoch = len(dataloader)
                    percent_complete = 100.0 * i / len(dataloader)
                    progress_message = (
                        f"[{num_samples}/{samples_per_epoch} ({percent_complete:.0f}%)]"
                    )
                except TypeError:
                    progress_message = f"[{num_samples} samples]"
                print(
                    f"Train Epoch: {epoch} {progress_message}\t"
                    f"Loss: {loss.item():.6f}\tData (t) {data_time:.3f}\tBatch (t) {batch_time:.3f}\t"
                    f"LR {optimizer.param_groups[0]['lr']:.5f}"
                )
    return model


def infer(model, dataloader, autocast, device):
    true, pred = [], []
    with torch.no_grad():
        for x, y in tqdm(dataloader):
            x = x.to(device)
            y = y.to(device)

            with autocast():
                logits = model(x)

            pred.append(logits.cpu())
            true.append(y.cpu())

    logits = torch.cat(pred)
    target = torch.cat(true)
    return logits, target


def find_peak(
    wd_list,
    idxs,
    train_loader,
    val_loader,
    input_shape,
    output_shape,
    lr,
    epochs,
    autocast,
    device,
    verbose,
    seed,
):
    best_wd_idx, max_acc = 0, 0
    for idx in idxs:
        weight_decay = wd_list[idx]
        model = train(
            train_loader,
            input_shape,
            output_shape,
            weight_decay,
            lr,
            epochs,
            autocast,
            device,
            seed,
        )
        logits, target = infer(model, val_loader, autocast, device)
        (acc1,) = accuracy(logits.float(), target.float(), topk=(1,))
        if verbose:
            print(f"Valid accuracy with weight_decay {weight_decay}: {acc1}")
        if max_acc < acc1:
            best_wd_idx, max_acc = idx, acc1
    return best_wd_idx


def evaluate(
    model,
    train_dataloader,
    dataloader,
    fewshot_k,
    batch_size,
    num_workers,
    lr,
    epochs,
    model_id,
    seed,
    feature_root,
    device,
    val_dataloader=None,
    normalize=True,
    amp=True,
    verbose=False,
):
    assert device == "cuda"  # need to use cuda for this else too slow
    # first we need to featurize the dataset, and store the result in feature_root
    if not os.path.exists(feature_root):
        os.mkdir(feature_root)
    feature_dir = os.path.join(feature_root, model_id)
    if not os.path.exists(feature_dir):
        os.mkdir(feature_dir)

    featurizer = Featurizer(model, normalize).cuda()
    autocast = torch.cuda.amp.autocast if amp else suppress
    if not os.path.exists(os.path.join(feature_dir, "targets_train.pt")):
        # now we have to cache the features
        devices = [x for x in range(torch.cuda.device_count())]
        featurizer = torch.nn.DataParallel(featurizer, device_ids=devices)

        splits = ["_train", "_val", "_test"]
        for save_str, loader in zip(
            splits, [train_dataloader, val_dataloader, dataloader]
        ):
            if loader is None:
                continue
            features = []
            targets = []
            num_batches_tracked = 0
            num_cached = 0
            with torch.no_grad():
                for images, target in tqdm(loader):
                    images = images.to(device)

                    with autocast():
                        feature = featurizer(images)

                    features.append(feature.cpu())
                    targets.append(target)

                    num_batches_tracked += 1
                    if (num_batches_tracked % 100) == 0:
                        features = torch.cat(features)
                        targets = torch.cat(targets)

                        torch.save(
                            features,
                            os.path.join(
                                feature_dir, f"features{save_str}_cache_{num_cached}.pt"
                            ),
                        )
                        torch.save(
                            targets,
                            os.path.join(
                                feature_dir, f"targets{save_str}_cache_{num_cached}.pt"
                            ),
                        )
                        num_cached += 1
                        features = []
                        targets = []

            if len(features) > 0:
                features = torch.cat(features)
                targets = torch.cat(targets)
                torch.save(
                    features,
                    os.path.join(
                        feature_dir, f"features{save_str}_cache_{num_cached}.pt"
                    ),
                )
                torch.save(
                    targets,
                    os.path.join(
                        feature_dir, f"targets{save_str}_cache_{num_cached}.pt"
                    ),
                )
                num_cached += 1

            features = torch.load(
                os.path.join(feature_dir, f"features{save_str}_cache_0.pt")
            )
            targets = torch.load(
                os.path.join(feature_dir, f"targets{save_str}_cache_0.pt")
            )
            for k in range(1, num_cached):
                next_features = torch.load(
                    os.path.join(feature_dir, f"features{save_str}_cache_{k}.pt")
                )
                next_targets = torch.load(
                    os.path.join(feature_dir, f"targets{save_str}_cache_{k}.pt")
                )
                features = torch.cat((features, next_features))
                targets = torch.cat((targets, next_targets))

            for k in range(num_cached):
                os.remove(os.path.join(feature_dir, f"features{save_str}_cache_{k}.pt"))
                os.remove(os.path.join(feature_dir, f"targets{save_str}_cache_{k}.pt"))

            torch.save(features, os.path.join(feature_dir, f"features{save_str}.pt"))
            torch.save(targets, os.path.join(feature_dir, f"targets{save_str}.pt"))

    features = torch.load(os.path.join(feature_dir, "features_train.pt"))
    targets = torch.load(os.path.join(feature_dir, "targets_train.pt"))

    # second, make a dataloader with k features per class. if k = -1, use all features.
    length = len(features)
    perm = [p.item() for p in torch.randperm(length)]
    idxs = []
    counts = {}
    num_classes = 0

    for p in perm:
        target = targets[p].item()
        if target not in counts:
            counts[target] = 0
            num_classes += 1

        if fewshot_k < 0 or counts[target] < fewshot_k:
            counts[target] += 1
            idxs.append(p)

    for c in counts:
        if fewshot_k > 0 and counts[c] != fewshot_k:
            print("insufficient data for this eval")
            return

    train_features = features[idxs]
    train_labels = targets[idxs]
    if val_dataloader is not None:
        features_val = torch.load(os.path.join(feature_dir, "features_val.pt"))
        targets_val = torch.load(os.path.join(feature_dir, "targets_val.pt"))
        feature_val_dset = FeatureDataset(features_val, targets_val)
        feature_val_loader = DataLoader(
            feature_val_dset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )
        feature_train_val_dset = FeatureDataset(
            np.concatenate((train_features, features_val)),
            np.concatenate((train_labels, targets_val)),
        )
        feature_train_val_loader = DataLoader(
            feature_train_val_dset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )
    feature_train_dset = FeatureDataset(train_features, train_labels)
    feature_train_loader = DataLoader(
        feature_train_dset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=True,
    )
    features_test = torch.load(os.path.join(feature_dir, "features_test.pt"))
    targets_test = torch.load(os.path.join(feature_dir, "targets_test.pt"))
    feature_test_dset = FeatureDataset(features_test, targets_test)
    feature_test_loader = DataLoader(
        feature_test_dset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=True,
    )
    input_shape, output_shape = features[0].shape[0], targets.max().item() + 1
    if val_dataloader is not None:
        # perform openAI-like hyperparameter sweep
        # https://arxiv.org/pdf/2103.00020.pdf A.3
        # instead of scikit-learn LBFGS use FCNNs with AdamW
        wd_list = np.logspace(-6, 2, num=97).tolist()
        wd_list_init = np.logspace(-6, 2, num=7).tolist()
        wd_init_idx = [i for i, val in enumerate(wd_list) if val in wd_list_init]
        peak_idx = find_peak(
            wd_list,
            wd_init_idx,
            feature_train_loader,
            feature_val_loader,
            input_shape,
            output_shape,
            lr,
            epochs,
            autocast,
            device,
            verbose,
            seed,
        )
        step_span = 8
        while step_span > 0:
            left, right = max(peak_idx - step_span, 0), min(
                peak_idx + step_span, len(wd_list) - 1
            )
            peak_idx = find_peak(
                wd_list,
                [left, peak_idx, right],
                feature_train_loader,
                feature_val_loader,
                input_shape,
                output_shape,
                lr,
                epochs,
                autocast,
                device,
                verbose,
                seed,
            )
            step_span //= 2
        best_wd = wd_list[peak_idx]
        train_loader = feature_train_val_loader
    else:
        best_wd = 0
        train_loader = feature_train_loader

    final_model = train(
        train_loader,
        input_shape,
        output_shape,
        best_wd,
        lr,
        epochs,
        autocast,
        device,
        seed,
    )
    logits, target = infer(final_model, feature_test_loader, autocast, device)
    pred = logits.argmax(axis=1)

    # measure accuracy
    if target.max() >= 5:
        acc1, acc5 = accuracy(logits.float(), target.float(), topk=(1, 5))
    else:
        (acc1,) = accuracy(logits.float(), target.float(), topk=(1,))
        acc5 = float("nan")
    mean_per_class_recall = balanced_accuracy_score(target, pred)
    fair_info = {
        "weight_decay": best_wd,
        "acc1": acc1,
        "acc5": acc5,
        "mean_per_class_recall": mean_per_class_recall,
        "classification_report": classification_report(target, pred, digits=3),
    }
    if verbose:
        print(fair_info["classification_report"])
        print(f"Test acc1: {acc1} with weight_decay: {best_wd}")
    return {
        "lp_acc1": fair_info["acc1"],
        "lp_acc5": fair_info["acc5"],
        "lp_mean_per_class_recall": fair_info["mean_per_class_recall"],
        "weight_decay": fair_info["weight_decay"],
        "epochs": epochs,
        "seed": seed,
        "fewshot_k": fewshot_k,
        "normalized": normalize,
    }



"""Tests for mc360iqa module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_mc360iqa_basics():
    from ayase.modules.mc360iqa import MC360IQAModule
    _test_module_basics(MC360IQAModule, "mc360iqa")

def test_mc360iqa_image(image_sample):
    from ayase.modules.mc360iqa import MC360IQAModule
    image_sample.quality_metrics = QualityMetrics()
    m = MC360IQAModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_mc360iqa_video(video_sample):
    from ayase.modules.mc360iqa import MC360IQAModule
    video_sample.quality_metrics = QualityMetrics()
    m = MC360IQAModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import torch
import torch.nn.functional as F
from ...common.registry import registry
from .blip import BlipBase
from torch import nn
from ..med import XBertEncoder

from ..vit import VisionTransformerEncoder


@registry.register_model("blip_image_text_matching")
class BlipITM(BlipBase):
    """
    BLIP Image-Text Matching (ITM) model.

    Supported model types:
        - base: fine-tuned BLIP retrieval weights on COCO dataset (Karpathy split).
        - large: fine-tuned BLIP retrieval weights on COCO dataset (Karpathy split).

    Usage:
        >>> from lavis.models import load_model
        >>> model = load_model("blip_image_text_matching", "base")
        >>> model = load_model("blip_image_text_matching", "large")
    """

    PRETRAINED_MODEL_CONFIG_DICT = {
        "base": "configs/models/blip_itm_base.yaml",
        "large": "configs/models/blip_itm_large.yaml",
    }

    def __init__(self, image_encoder, text_encoder, embed_dim=256, max_txt_len=35):
        super().__init__()

        self.tokenizer = self.init_tokenizer()

        self.text_encoder = text_encoder

        self.visual_encoder = image_encoder

        self.max_txt_len = max_txt_len

        # creating projection layers for ITC
        text_width = text_encoder.config.hidden_size
        vision_width = image_encoder.vision_width

        self.vision_proj = nn.Linear(vision_width, embed_dim)
        self.text_proj = nn.Linear(text_width, embed_dim)

        self.itm_head = nn.Linear(text_width, 2)

    def forward(self, samples, match_head="itm"):
        image = samples["image"]
        caption = samples["text_input"]

        image_embeds = self.visual_encoder.forward_features(image)
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            image.device
        )

        text = self.tokenizer(
            caption,
            padding="longest",
            truncation=True,
            max_length=self.max_txt_len,
            return_tensors="pt",
        ).to(image.device)
        if match_head == "itm":
            encoder_input_ids = text.input_ids.clone()
            encoder_input_ids[:, 0] = self.tokenizer.enc_token_id  # extra code
            output = self.text_encoder(
                encoder_input_ids,
                attention_mask=text.attention_mask,
                encoder_hidden_states=image_embeds,
                encoder_attention_mask=image_atts,
                return_dict=True,
            )
            itm_output = self.itm_head(output.last_hidden_state[:, 0, :])
            return itm_output

        elif match_head == "itc":
            text_output = self.text_encoder(
                text.input_ids,
                attention_mask=text.attention_mask,
                return_dict=True,
                mode="text",
            )
            image_feat = F.normalize(self.vision_proj(image_embeds[:, 0, :]), dim=-1)
            text_feat = F.normalize(
                self.text_proj(text_output.last_hidden_state[:, 0, :]), dim=-1
            )

            sim = image_feat @ text_feat.t()
            return sim
    def itm_rank(self, image_embeds, image_atts, encoder_input_ids, match_head='itm'):
        # breakpoint()
        encoder_input_ids = encoder_input_ids.clone()
        encoder_input_ids = encoder_input_ids[:, 3:]
        text_attention_mask = (encoder_input_ids != self.tokenizer.pad_token_id).long()

        if match_head == 'itm':
            # encoder_input_ids = encoder_input_ids.clone()
            encoder_input_ids[:, 0] = self.tokenizer.enc_token_id
            output = self.text_encoder(encoder_input_ids,
                                       attention_mask=text_attention_mask,
                                       encoder_hidden_states=image_embeds,
                                       encoder_attention_mask=image_atts,
                                       return_dict=True,
                                       )
            # print(output.last_hidden_state.shape)
            itm_output = self.itm_head(output.last_hidden_state[:, 0, :])
            itm_output = F.softmax(itm_output, dim=1)[:,1]
            return itm_output #, mask, token_length

        elif match_head == 'itc':
            encoder_input_ids[:, 0] = self.tokenizer.cls_token_id
            text_output = self.text_encoder(encoder_input_ids, attention_mask=text_attention_mask,
                                            return_dict=True, mode='text')
            image_feat = F.normalize(self.vision_proj(image_embeds[:, 0, :]), dim=-1)
            text_feat = F.normalize(self.text_proj(text_output.last_hidden_state[:, 0, :]), dim=-1)

            sim = image_feat @ text_feat.t()
            return sim

    @classmethod
    def from_config(cls, cfg=None):
        image_encoder = VisionTransformerEncoder.from_config(cfg)
        text_encoder = XBertEncoder.from_config(cfg)

        embed_dim = cfg.get("embed_dim", 256)
        max_txt_len = cfg.get("max_txt_len", 35)

        model = cls(
            image_encoder=image_encoder,
            text_encoder=text_encoder,
            embed_dim=embed_dim,
            max_txt_len=max_txt_len,
        )

        model.load_checkpoint_from_config(cfg)

        return model


def compute_gradcam(model, visual_input, text_input, tokenized_text, block_num=6):
    model.text_encoder.base_model.base_model.encoder.layer[
        block_num
    ].crossattention.self.save_attention = True

    output = model({"image": visual_input, "text_input": text_input}, match_head="itm")
    loss = output[:, 1].sum()

    model.zero_grad()
    loss.backward()
    with torch.no_grad():
        mask = tokenized_text.attention_mask.view(
            tokenized_text.attention_mask.size(0), 1, -1, 1, 1
        )  # (bsz,1,token_len, 1,1)
        token_length = tokenized_text.attention_mask.sum(dim=-1) - 2
        token_length = token_length.cpu()
        # grads and cams [bsz, num_head, seq_len, image_patch]
        grads = model.text_encoder.base_model.base_model.encoder.layer[
            block_num
        ].crossattention.self.get_attn_gradients()
        cams = model.text_encoder.base_model.base_model.encoder.layer[
            block_num
        ].crossattention.self.get_attention_map()

        # assume using vit with 576 num image patch
        cams = cams[:, :, :, 1:].reshape(visual_input.size(0), 12, -1, 24, 24) * mask
        grads = (
            grads[:, :, :, 1:].clamp(0).reshape(visual_input.size(0), 12, -1, 24, 24)
            * mask
        )

        gradcams = cams * grads
        gradcam_list = []

        for ind in range(visual_input.size(0)):
            token_length_ = token_length[ind]
            gradcam = gradcams[ind].mean(0).cpu().detach()
            # [enc token gradcam, average gradcam across token, gradcam for individual token]
            gradcam = torch.cat(
                (
                    gradcam[0:1, :],
                    gradcam[1 : token_length_ + 1, :].sum(dim=0, keepdim=True)
                    / token_length_,
                    gradcam[1:, :],
                )
            )
            gradcam_list.append(gradcam)
            
    return gradcam_list, output
